/* library functions for spherical astronomy 
   Thomas Szeifert - LSW - Heidelberg - 11.12.1997


   precession: compute precesssion from one to next epoche 
   transtheta: compute the rotation offset of a vector on the sphere for 
               precession from one to an other epoche - Call first precession
	       to get earth precession constant "the"
   compst:     compute local siderial time	       
   hor2equ:    compute horizon to equatorial coordinates
   equ2hor:    compute equatorial to horizon coordinates

   dayofyear:  civil date to day of the year
   juldat(double date[],double ut, double *jd):
               compute Julian day 
   gal2equ(double l, double b, double *a, double *d)
               galactic to equatorial coordinate
   equ2gal(double a, double d, double *l, double *b)
               equatorial to galactic coordinate
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <proto_fors.h>

#ifndef PI
#define PI (acos((double) -1.0))
#endif
#define RAD (((double) 180.0) / PI)
#define HRAD (((double) 12.0) / PI)


/* explanatory supplement to the astronomical almanac 1992 p104 */
#ifdef __STDC__
void precession(double a,double d, double e[], double *ac, double *dc, 
		double *zet, double *z, double *the)
#else
void precession(a, d, e, ac, dc, zet, z, the)
double a, d, e[], *ac, *dc, *zet, *z, *the;
#endif
{

  double zeti[3][3],zi[3][3],thei[3][3];
  double tmp, tmp1;
  double ec,t,tg;
  double ra[3],dec[3];
  
  int i,j;


  for (i = 0 ; i < 3 ; i++)
    for (j = 0 ; j < 3 ; j++)
      zeti[i][j] = zi[i][j] = thei[i][j] = (double) 0.0;

  /* explanatory supplement to the astronomical almanac 1992 : 
     precission constants from year e[0] toward year e[1]*/
  zeti[0][0] = (double) 2306.2181 , zeti[1][0] = (double) 1.39656,
    zeti[2][0] = (double) -0.000139;
  zeti[0][1] = (double) 0.30188, zeti[1][1] = (double) -0.000344;
  zeti[0][2] = (double) 0.017998;
  zi[0][0] = (double) 2306.2181 , zi[1][0] = (double) 1.39656,
    zi[2][0] = (double) -0.000139;
  zi[0][1] = (double) 1.09468, zi[1][1] = (double) 0.000066;
  zi[0][2] = (double) 0.018203;
  thei[0][0] = (double) 2004.3109 , thei[1][0] = (double) -0.85330,
    thei[2][0] = (double) -0.000217;
  thei[0][1] = (double) -0.42665, thei[1][1] = (double) -0.000217;
  thei[0][2] = (double) -0.041833;

  for (i = 0 ; i < 3 ; i++)
    for (j = 0 ; j < 3 ; j++)
      {
	zeti[i][j] /= (double) 3600.0;
	zi[i][j] /= (double) 3600.0;
	thei[i][j] /= (double) 3600.0;
      }

  a = a / HRAD;
  d = d / RAD;

  ec = (double) 2000.0;

      

  /* explanatory supplement to the astronomical almanac 1998 :
     TRASFFORMATION FROM INITIAL TO FINAL epoche */

  t = (e[0] - e[1]) / (double) 100.0; 
  tg = (e[1] - ec) / (double) 100.0; 

  *zet = *z = *the = (double) 0.0;

  for ( i = 0 ; i < 3 ; i++)
    for ( j = 0 ; j < 3 ; j++)
      {
	*zet += (zeti[j][i] * pow(t,(double) (i+1)) * pow(tg,(double) (j)));
	*z += (zi[j][i] * pow(t,(double) i+1) * pow(tg,(double) (j)));
	*the += (thei[j][i] * pow(t,(double) (i+1)) * pow(tg,(double) (j)));
      }


  *zet = *zet / RAD;
  *z = *z / RAD;
  *the = *the / RAD;

  /*  
  printf("\n     zet = %f z = %f the = %f deg\n",RAD * *zet,RAD * *z,RAD * *the);

  printf("\n     zet = %22.16e\n     z   = %22.16e\n     the = %22.16e rad\n",
	 *zet,*z,*the);
  */

  *dc = asin(((double) -1.0)*cos(a-*z)*sin(*the)*cos(d) + cos(*the)*sin(d));

  tmp  = sin(a-*z)*cos(d);
  tmp1 = cos(a-*z)*cos(*the)*cos(d)+sin(*the)*sin(d);


  *ac = atan2(tmp,tmp1) - *zet;


  if (*ac < 0) *ac += (2.0 * PI);
  if (*ac > 2.0 * PI) *ac -= (2.0 * PI);

  /*
  printf("\nRA, DEC of epoche %f = %f, %f", e[1], *ac * HRAD, *dc * RAD); 
  */

  tmp = *ac * HRAD;
  ra[0] = floor(tmp);
  tmp = (*ac * HRAD - ra[0]) * (double) 60.0;
  ra[1] = floor(tmp);
  ra[2] = (tmp - ra[1]) * (double) 60.0;


  tmp = *dc * RAD;
  if (tmp > 0.0){
    dec[0] = floor(tmp);
    tmp = (*dc * RAD - dec[0]) * (double) 60.0;
    dec[1] = floor(tmp);
    dec[2] = (tmp - dec[1]) * (double) 60.0;
  }
  else{
    dec[0] = floor(tmp+1.0);
    tmp = (*dc * RAD - dec[0]) * (double) 60.0;
    dec[1] = fabs(floor(tmp)+1.0);
    dec[2] = fabs((tmp + dec[1]) * (double) 60.0);
  }

  /*  
  if (d < 0 && dec[0] == 0)
    printf("RA, DEC of epoche %f = %2.0f,%2.0f,%6.3f, -%2.0f,%2.0f,%5.2f",e[1],ra[0],ra[1],ra[2],dec[0],dec[1],dec[2]); 
  else
    printf("RA, DEC of epoche %f = %2.0f,%2.0f,%6.3f, %3.0f,%2.0f,%5.2f",e[1],ra[0],ra[1],ra[2],dec[0],dec[1],dec[2]); 
    */

  if (d < 0 && dec[0] == 0)
    printf("(%6.1f) = %2.0f,%2.0f,%6.3f  -%2.0f,%2.0f,%5.2f",e[1],ra[0],ra[1],ra[2],dec[0],dec[1],dec[2]); 
  else
    printf("(%6.1f) = %2.0f,%2.0f,%6.3f  %3.0f,%2.0f,%5.2f",e[1],ra[0],ra[1],ra[2],dec[0],dec[1],dec[2]); 

  *ac *= HRAD;
  *dc *= RAD;

  

  return;
}






/* Newcomb: spherical astronomy 1906, p272 */
#ifdef __STDC__
void transtheta(double a, double d, double the, double *thetac)
#else
void transtheta(a, d, the, thetac)
double a, d, the, *thetac;
#endif
{
  double tmp;

  a /= HRAD;
  d /= RAD;

  tmp = (sin(the) * sin(a)) / cos(d);

  *thetac = asin(tmp);

  *thetac *= RAD;

  return;
}


/* Astronomical almanac 1998 B6 -- equation of equinox was neglected*/
#ifdef __STDC__
void compst(double day, double ut, double phi, double lam, double *gmst, 
	    double *st, char mode)
#else
void compst(day, ut, phi, lam, gmst, st,mode)
double day,ut,phi,lam,*gmst,*st;
char mode;
#endif
{
  double t,c0[4],r,c1[4],ec,tmp;
  int i;

  c0[0] = (double) 24110.54841, c0[1] = (double) 8640184.812866;
  c0[2] = (double) 0.093104,    c0[3] = (double) -6.2e-06;

  c1[0] = 1.002737909350795, c1[1] = 5.9006e-11, c1[2] = 5.9e-15, c1[3] = 0.0;

  if (toupper(mode) != 'J'){
    ec = 2000.0; /* case civil year */
    t = (day - ec) / (double) 100.0;
  }
  else{
    ec = 2451545.0; /* case Julian day */
    t = (day - ec) / (double) 36525.0;
  }

  *gmst = 0.0;
  r = (double) 0.0;

  for (i = 0 ; i < 4 ; i++){
    *gmst += c0[i] * pow(t, ((double) i));
    r += c1[i] * pow(t, ((double) i));
  }


  *gmst = 24.0*(*gmst/(3600.0*24.0) - floor(*gmst/(3600.0*24.0)));
  
  /*
  printf ("r = %f GMST = %f \n",r,*gmst);
  */

  tmp = *gmst + lam + ut*r;
  tmp = (tmp > 0.0) ? tmp : tmp + 24.0;
  tmp = (tmp < 24.0) ? tmp : tmp - 24.0;

  *st = tmp;

  return;
}


/* Duffet-Smith 1981 p40*/
/*Yegoro and Sabinin, 1971; in New Techniques in Astronomy, p 108, Edt Ingrao,
Gordon and Breach science publisher*/
#ifdef __STDC__
void equ2hor(double h, double d, double phi, double *al, double *az, 
	     double *q)
#else
void equ2hor(h,d,phi,al,az,q)
     double h, d, phi, *al, *az, *q;
#endif
{
  double tmp,tmp1,tmp2,tmp3;

      d /= RAD;
      h /= HRAD;

      /* Duffet-Smith 1981 p40*/
      *al = asin(sin(d)*sin(phi) + cos(d)*cos(phi)*cos(h));
      tmp = (sin(d) - sin(phi)*sin(*al))/(cos(phi)*cos(*al));
      *az = (tmp > -1.0) ? acos(tmp) : PI;

      *az = (sin(h) < 0.0) ? *az : ((double) 2.0) * PI - *az;
      *az = (*az >= ((double) 2.0) * PI) ? *az - ((double) 2.0) * PI : *az;
      

      /*Newcomb 1906 p133:*/
      tmp1 = cos(phi)*sin(h);
      tmp2 = sin(phi)*cos(h) - cos(phi)*sin(d)*cos(h);

      tmp3 = RAD * atan2(tmp1,tmp2);

      *q = (tmp3 < 0.0) ? ((double) 360) + tmp3 : tmp3;

      *al *= RAD;
      *az *= RAD;



      return;
}





/* Duffet-Smith 1981 p40*/
#ifdef __STDC__
void hor2equ(double az, double al, double phi, double *h, double *d, 
             double *q)
#else
void hor2equ(az,al,phi,h,d,q)
     double al, az, phi, *h, *d, *q;
#endif
{
  double tmp,tmp1,tmp2,tmp3;

      al = al / RAD;
      az = az / RAD;
    
      *d = asin( sin(al) * sin(phi) + cos(al) * cos(phi) * cos(az));

      tmp = (sin(al) - sin(phi) * sin(*d))/(cos(phi) * cos(*d));

      /* printf ("tmp = %f\n",tmp);*/

      /* tmp can be < -1 because of numerical problems in the previous 
	 calculation (at az = 0.0) */
      *h = (tmp > -1.0) ? acos(tmp) : PI;
      *h = (sin(az) < 0.0) ? *h : ((double)2) * PI - *h;
      *h = (*h >= ((double) 2) * PI ) ? *h - ((double) 2) * PI : *h;


      /*Newcomb 1906 p133:*/
      tmp1 = cos(phi)*sin(*h);
      tmp2 = sin(phi)*cos(*h) - cos(phi)*sin(*d)*cos(*h);

      tmp3 = RAD * atan2(tmp1,tmp2);

      *q = (tmp3 < 0.0) ? ((double) 360) + tmp3 : tmp3;
      *q = (*q >= 360.0) ? *q - ((double) 360) : *q;


      *d *= RAD;
      *h *= HRAD;



      return;
}


/* Meeus 1982 p28 */
#ifdef __STDC__
int dayofyear(double date[])
#else
int dayofyear(date)
     double date[];
#endif
{
  int idate[3];
  int i,itmp=0;

  for (i = 0 ; i < 3 ; i++)
    idate[i] = (int) date[i];

  /*WARNING: we really want to compute integers without considering the
    modulo of the fraction!!!*/


  /* leap years: */
  if (((idate[0] % 4) == 0 && (idate[0] % 100) != 0) || (idate[0] % 400) == 0){
    itmp = ((275*idate[1])/9 - (idate[1]+9)/12) + idate[2] - 30;
  }
  else{
    itmp = ((275*idate[1])/9 - 2*(idate[1]+9)/12) + idate[2] - 30;
  }

  printf ("it's the %d day of the year\n",itmp);

  return (itmp);
}


/* Meeus 1982 p28 */
#ifdef __STDC__
void juldat(double date[],double ut, double *jd)
#else
void juldat(date,ut,jd)
     double date[],ut,*jd;
#endif
{
  int idate[3],i,ia,ib;
  double frac,yp,p,c,a,nd;




  if ( date[0] - floor(date[0]) > 1.0e-6){

    if (date[1] > 1.0e-6 || date[2] > 1.0e-6){
      printf("Error: Date was entered in wrong format.\n");
      SCETER(1,"Error: Date was entered in wrong format.\n");
    }
    /*FIRST CASE: format was year.fraction*/
    idate[0] = (int) date[0]; 
    frac = date[0] - floor(date[0]);
    idate[1] = idate[2] = 1;
    date[1] = date[2] = 1.0;
  }
  else{
    for (i = 0 ; i < 3 ; i++)
      idate[i] = NINT(date[i]);
    frac = 0.0;
    if (idate[1] == 0 && idate[2] == 0){
      idate[1] = idate[2] = 1;
      date[1] = date[2] = 1.0;
    }
    if (idate[1] <= 0 && idate[1] > 12)
      SCETER(1,"Error: such a month does not exist");
    if (idate[2] <= 0 && idate[2] > 31)
      SCETER(1,"Error: such a day does not exist");
  }

  if (idate[1] > 2){
    yp = (double) idate[0]; 
    p = (double) idate[1];
  }
  else{
    yp = (double) (idate[0]-1); 
    p = (double) (idate[1]+12);
  }

  c = date[0] + date[1] / 100.0 + date[2] / 10000.0 + ut / 1000000.0;
  if (c >=  1582.1015){ /*Gregorian date*/
    ia = (int) (yp/100.0);
    a = (double) ia;
    ib = 2 - ia + (int) (a/4.0);
  }
  else{
    ib = 0;
  }

  *jd = floor(365.25*yp) + floor(30.6001*(p+1.0)) + date[2] + ((double) ib) + 
    1720994.5;

  if (frac > 1.0e-6){

    if (( idate[0] % 4 == 0 && idate[0] % 100 != 0) || idate[0] % 400 == 0)
      nd = 366.0;
    else
      nd = 365.0;
   
    if (fabs( frac*nd - (double) (NINT(frac*nd))) > 0.1)
      printf("Warning: Fraction of year MAY not correspond to integer number of days\n");

    printf("%f %f %d %f\n",*jd, frac, NINT(frac * nd), (double) (NINT(frac * nd)));

    *jd = *jd + ((double) NINT(frac * nd)) - 0.5;
  }
  


  return;
}






/* Duffet-Smith 1981 p50*/
#ifdef __STDC__
void gal2equ(double l, double b, double *a, double *d)
#else
void gal2equ(l,b,a,d)
     double l, b, *a, *d;
#endif
{
  double tmp,tmp1,tmp2,tmp3;
  double gca=192.25,gcd=27.4,angp=33.0;

  gca /= 15.0;
  gcd /= RAD;
  angp /= RAD;

  l /= RAD;
  b /= RAD;
    

  *d = asin( cos(b) * sin(l-angp) * cos(gcd) + sin(b) * sin(gcd));
  *d *= RAD;


  tmp1 = cos(b) * cos(l - angp); 
  tmp2 = sin(b) * cos(gcd) - cos(b) * sin(gcd) * sin(l - angp);

  tmp3 = HRAD * atan2(tmp1,tmp2);
  tmp3 += gca;

  *a = (tmp3 < 0.0) ? 24.0 + tmp3 : tmp3;



  return;
}


/* Duffet-Smith 1981 p48*/
#ifdef __STDC__
void equ2gal(double a, double d, double *l, double *b)
#else
void equ2gal(a,d,l,b)
     double a, d, *l, *b;
#endif
{
  double tmp,tmp1,tmp2,tmp3;
  double gca=192.25,gcd=27.4,angp=33.0;

  gca /= RAD;
  gcd /= RAD;

  d = d / RAD;
  a = a / HRAD;


  *b = asin(cos(d) * cos(gcd) * cos(a-gca) + sin(d) * sin(gcd));


  tmp1 = sin(d) - sin(*b) * sin(gcd);
  tmp2 = cos(d) * sin(a-gca) * cos(gcd);

  tmp3 = RAD * atan2(tmp1,tmp2);
  tmp3 += angp;

  *l = (tmp3 < 0.0) ? 360.0 + tmp3 : tmp3;
  *l = (*l >= 360.0) ? *l - 360 : *l;

  *b *= RAD;

  return;
}







