/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.IDENT        fitnol.c
.MODULE       subroutines 
.LANGUAGE     C
.AUTHOR       Cristian Levin - ESO La Silla
.AUTHOR       Thomas Szeifert - LSW Heidelberg
.PURPOSE      1D- and 2D-fitting.
.KEYWORDS     fitting.
.COMMENTS     Most of this routines were taken as they are from
              the book "Numerical Recipes in C" -- 1st edition
	      now we use the original numerical recipes (but double) 
	      Some changes by Christian were eliminated recently.
	      It's now possible to fix parameters if sig[i] = 0
.VERSION      Version March 1997  
.ENVIRONMENT  UNIX
------------------------------------------------------------*/


/*
  functions of fitnol.c


  void mrqmin(double x[], double y[], double sig[],int ndata,double a[],
    int ia[], int ma, double **covar, double **alpha, double *chisq,
    void (*funcs)(double,double [],double *,double [],int),double *alamda)



  void mrqcof(double x[],double y[],double sig[],int ndata,double a[],int ia[],
    int ma, double **alpha, double beta[], double *chisq,
    void (*funcs)(double, double [], double *, double [],int))



  void fgauss(double x, double a[], double *y, double dyda[],int inull)



  fit_gauss( double *x, double *y, int n, double *a )




  void mrqmin2D(double x1[],double x2[],double y[],double sig[],int ndata,
    double a[],int ia[],int ma,double **covar,double **alpha,double *chisq,
    void (*funcs)(double,double,double [],double *,double [],int),
    double *alamda)



  void mrqcof2D(double x1[],double x2[],double y[],double sig[],int ndata,
    double a[],int ia[], int ma, double **alpha, double beta[], double *chisq,
    void (*funcs)(double, double, double [], double *, double [],int))
 */


#include <math.h>

#include <proto_fors.h>

int inull = 0; /* here for the Gaussian we don't need the number in funcs*/



#ifdef __STDC__
void mrqmin(double x[], double y[], double sig[],int ndata,double a[],int ia[],
	int ma, double **covar, double **alpha, double *chisq,
	void (*funcs)(double,double [],double *,double [],int),double *alamda)
#else
void mrqmin( x, y,  sig,  ndata,  a, ia,
	ma,  covar,  alpha,  chisq,  funcs,  alamda)

  double x[],y[],sig[], a[], **covar, **alpha,  *chisq, *alamda;
  int  ndata, ia[],ma;
  void   (*funcs) ();
#endif
{
	int j,k,l,m;
	static int mfit;
	static double ochisq,*atry,*beta,*da,**oneda;

	if (*alamda < 0.0) {
		atry=dvector(1,ma);
		beta=dvector(1,ma);
		da=dvector(1,ma);
		for (mfit=0,j=1;j<=ma;j++)
			if (ia[j]) mfit++;
		oneda=dmatrix(1,mfit,1,1);
		*alamda=0.001;
		mrqcof(x,y,sig,ndata,a,ia,ma,alpha,beta,chisq,funcs);
		ochisq=(*chisq);
		for (j=1;j<=ma;j++) atry[j]=a[j];
	}
	for (j=0,l=1;l<=ma;l++) {
		if (ia[l]) {
			for (j++,k=0,m=1;m<=ma;m++) {
				if (ia[m]) {
					k++;
					covar[j][k]=alpha[j][k];
				}
			}
			covar[j][j]=alpha[j][j]*(1.0+(*alamda));
			oneda[j][1]=beta[j];
		}
	}
	dgaussj(covar,mfit,oneda,1);
	for (j=1;j<=mfit;j++) da[j]=oneda[j][1];
	if (*alamda == 0.0) {
		covsrt(covar,ma,ia,mfit);
		free_dmatrix(oneda,1,mfit,1,1);
		free_dvector(da,1,ma);
		free_dvector(beta,1,ma);
		free_dvector(atry,1,ma);
		return;
	}
	for (j=0,l=1;l<=ma;l++)
		if (ia[l]) atry[l]=a[l]+da[++j];
	mrqcof(x,y,sig,ndata,atry,ia,ma,covar,da,chisq,funcs);
	if (*chisq < ochisq) {
		*alamda *= 0.1;
		ochisq=(*chisq);
		for (j=0,l=1;l<=ma;l++) {
			if (ia[l]) {
				for (j++,k=0,m=1;m<=ma;m++) {
					if (ia[m]) {
						k++;
						alpha[j][k]=covar[j][k];
					}
				}
				beta[j]=da[j];
				a[l]=atry[l];
			}
		}
	} else {
		*alamda *= 10.0;
		*chisq=ochisq;
	}
}

#ifdef __STDC__
void mrqcof(double x[],double y[],double sig[],int ndata,double a[],int ia[],
	int ma, double **alpha, double beta[], double *chisq,
	void (*funcs)(double, double [], double *, double [],int))
#else
void mrqcof(x, y, sig, ndata, a, ia, ma,  alpha,  beta,  chisq, funcs)  
  double x[],y[],sig[],a[],**alpha,beta[],*chisq;
  int ndata,ma,ia[];
  void (*funcs)();
#endif
{
	int i,j,k,l,m,mfit=0;
	double ymod,wt,sig2i,dy,*dyda;

	dyda=dvector(1,ma);
	for (j=1;j<=ma;j++)
		if (ia[j]) mfit++;
	for (j=1;j<=mfit;j++) {
		for (k=1;k<=j;k++) alpha[j][k]=0.0;
		beta[j]=0.0;
	}
	*chisq=0.0;
	for (i=1;i<=ndata;i++) {
		(*funcs)(x[i],a,&ymod,dyda,ma);
		sig2i=1.0/(sig[i]*sig[i]);
		dy=y[i]-ymod;
		for (j=0,l=1;l<=ma;l++) {
			if (ia[l]) {
				wt=dyda[l]*sig2i;
				for (j++,k=0,m=1;m<=l;m++)
					if (ia[m]) alpha[j][++k] += wt*dyda[m];
				beta[j] += dy*wt;
			}
		}
		*chisq += dy*dy*sig2i;
	}
	for (j=2;j<=mfit;j++)
		for (k=1;k<j;k++) alpha[k][j]=alpha[j][k];
	free_dvector(dyda,1,ma);
}


#define LSWAP(a,b) {swap=(a);(a)=(b);(b)=swap;}

#ifdef __STDC__
void covsrt(double **covar, int ma, int ia[], int mfit)
#else
void covsrt( covar,  ma,  ia, mfit)
  double **covar;
  int  ma, ia[], mfit;
#endif
{
	int i,j,k;
	double swap;

	for (i=mfit+1;i<=ma;i++)
		for (j=1;j<=i;j++) covar[i][j]=covar[j][i]=0.0;
	k=mfit;
	for (j=ma;j>=1;j--) {
		if (ia[j]) {
			for (i=1;i<=ma;i++) LSWAP(covar[i][k],covar[i][j])
			for (i=1;i<=ma;i++) LSWAP(covar[k][i],covar[j][i])
			k--;
		}
	}
}

#undef LSWAP






/************************************************************
  fgauss(): optimized adding fac1, fac2. (C.Levin)
  	    optimized using only 3 coefs. (1 gaussian) (C.Levin).
*/
#ifdef __STDC__
   void fgauss(double x, double a[], double *y, double dyda[],int inull)
#else
   void fgauss( x, a, y, dyda,inull)
   double x,a[],*y,dyda[];
   int inull;
#endif
{

	double fac1, fac2, ex, arg;

	*y = 0.0;
	
	arg = (x - a[2]) / a[3];
	dyda[1] = ex = exp(-0.5 * arg * arg);
 	*y = fac1 = a[1] * ex;
	dyda[2] = fac2 = fac1 * 2.0 * arg / a[3];
	dyda[3] = fac2 * arg;
}

/************************************************************
 *
 * fit_gauss(): Gaussian fitting. 
 * 
 * calls   : fitnol.c{mrqmin} 
 * modified: Criterium of stopping is more relaxed (C.Levin).
 *
 ************************************************************/


#define EPS	0.001       /* delta chi**2 / chi**2    TS-0896*/
#define NITER   20          /* max number of iteration*/

#ifdef __STDC__
int
   fit_gauss( double *x, double *y, int n, double *a )
#else
int
   fit_gauss( x, y, n, a )
   double *x,*y,*a;
   int n;
#endif
{
    int *ia;
    int nfit = 3, ncoefs = 3;
    int i, iter = 1;
    double **covar, **alpha;
    double *sig, chisq, ochisq, alamda = -1.;

    sig = dvector( 1, n );

    ia = ivector(1, 3);
    covar = dmatrix( 1, 3, 1, 3 );
    alpha = dmatrix( 1, 3, 1, 3 );

    /*
    for ( i = 1; i <= n; i++ )
	sig[i] = 1.0;*/

    for ( i = 1; i <= n; i++ )
      sig[i] = (y[i] > 0) ? sqrt(y[i]) : 1.0;


    for ( i = 1; i <= 3; i++ )
	ia[i] = 1;


    mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
	    &alamda );

    do {
	iter++;
	ochisq = chisq;
	mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
		&alamda );
    } while ( (iter < NITER) && 
	      (((ochisq - chisq) / chisq > EPS) || (ochisq ==  chisq)) 
	      ); 
    /*} while (((ochisq - chisq) / chisq > EPS) || (iter <= NITER));*/ 
    /*limit niter by TS0896*/

    /*
    if (iter == NITER && ((ochisq - chisq) / chisq > EPS))
      {
	printf("mrqmin didn't reach convergence -- eps = %11.5f \n",
	       (ochisq - chisq) / chisq);
      }
    if (iter == NITER && (ochisq == chisq))
      {
	printf("mrqmin didn't find a minimum solution\n");
      }
      */

    alamda = 0.;  /* To de-allocate memory */

    mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
	    &alamda );

    free_dvector( sig, 1, n );
    free_ivector( ia, 1, 3 );
    free_dmatrix( covar, 1, 3, 1, 3 );
    free_dmatrix( alpha, 1, 3, 1, 3 );

    return iter;
}

#undef EPS
#undef NITER



/************************************************************
 *
 * fit_gauss_weight(): Gaussian fitting. 
 * 
 * calls   :  fitnol.c {mrqmin} 
 * modified:  Criterium of stopping is more relaxed (C.Levin).
 * modified:  take into account that you don't have equal weights
 *            allow to freeze parameters (Thomas Szeifert 15.7.98)
 * 
 ************************************************************/


#define EPS	0.001       /* delta chi**2 / chi**2    TS-0896*/
#define NITER   20          /* max number of iteration*/

#ifdef __STDC__
int
fit_gauss_weight (double *x, double *y, double *sig, int n, double *a, 
		     int *ia)
#else
int
fit_gauss_weight( x, y, sig, n, a, ia )
     double *x,*y,*a,*sig;
     int n, ia;
#endif
{
    int nfit = 3, ncoefs = 3;
    int i, iter = 1;
    double **covar, **alpha;
    double chisq, ochisq, alamda = -1.;


    covar = dmatrix( 1, 3, 1, 3 );
    alpha = dmatrix( 1, 3, 1, 3 );



    mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
	    &alamda );

    do {
	iter++;
	ochisq = chisq;
	mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
		&alamda );
    } while ( (iter < NITER) && 
	      (((ochisq - chisq) / chisq > EPS) || (ochisq ==  chisq)) 
	      ); /* limit niter by TS0896 */

    /*
    if (iter == NITER && ((ochisq - chisq) / chisq > EPS))
      {
	printf("mrqmin didn't reach convergence -- eps = %11.5f \n",
	       (ochisq - chisq) / chisq);
      }
    if (iter == NITER && (ochisq == chisq))
      {
	printf("mrqmin didn't find a minimum solution\n");
      }
      */

    alamda = 0.;  /* To de-allocate memory */

    mrqmin( x, y, sig, n, a, ia, ncoefs, covar, alpha, &chisq, fgauss, 
	    &alamda );

    free_dmatrix( covar, 1, 3, 1, 3 );
    free_dmatrix( alpha, 1, 3, 1, 3 );

    return iter;
}

#undef EPS
#undef NITER





#ifdef __STDC__
void mrqmin2D(double x1[],double x2[],double y[],double sig[],int ndata,
	double a[],int ia[],int ma,double **covar,double **alpha,double *chisq,
	void (*funcs)(double,double,double [],double *,double [],int),
	      double *alamda)
#else
void mrqmin2D( x1, x2, y,  sig,  ndata,  a, ia,
	ma,  covar,  alpha,  chisq,  funcs, alamda)

  double x1[],x2[],y[],sig[], a[], **covar, **alpha,  *chisq,*alamda;
  int  ndata, ia[],ma;
  void   (*funcs) ();
#endif
{
	int j,k,l,m;
	static int mfit;
	static double ochisq,*atry,*beta,*da,**oneda;

	if (*alamda < 0.0) {
		atry=dvector(1,ma);
		beta=dvector(1,ma);
		da=dvector(1,ma);
		for (mfit=0,j=1;j<=ma;j++)
			if (ia[j]) mfit++;
		oneda=dmatrix(1,mfit,1,1);
		*alamda=0.001;
		mrqcof2D(x1,x2,y,sig,ndata,a,ia,ma,alpha,beta,chisq,funcs);
		ochisq=(*chisq);
		for (j=1;j<=ma;j++) atry[j]=a[j];
	}
	for (j=0,l=1;l<=ma;l++) {
		if (ia[l]) {
			for (j++,k=0,m=1;m<=ma;m++) {
				if (ia[m]) {
					k++;
					covar[j][k]=alpha[j][k];
				}
			}
			covar[j][j]=alpha[j][j]*(1.0+(*alamda));
			oneda[j][1]=beta[j];
		}
	}
	dgaussj(covar,mfit,oneda,1);
	for (j=1;j<=mfit;j++) da[j]=oneda[j][1];
	if (*alamda == 0.0) {
		covsrt(covar,ma,ia,mfit);
		free_dmatrix(oneda,1,mfit,1,1);
		free_dvector(da,1,ma);
		free_dvector(beta,1,ma);
		free_dvector(atry,1,ma);
		return;
	}
	for (j=0,l=1;l<=ma;l++)
		if (ia[l]) atry[l]=a[l]+da[++j];
	mrqcof2D(x1,x2,y,sig,ndata,atry,ia,ma,covar,da,chisq,funcs);
	if (*chisq < ochisq) {
		*alamda *= 0.1;
		ochisq=(*chisq);
		for (j=0,l=1;l<=ma;l++) {
			if (ia[l]) {
				for (j++,k=0,m=1;m<=ma;m++) {
					if (ia[m]) {
						k++;
						alpha[j][k]=covar[j][k];
					}
				}
				beta[j]=da[j];
				a[l]=atry[l];
			}
		}
	} else {
		*alamda *= 10.0;
		*chisq=ochisq;
	}
}

#ifdef __STDC__
void mrqcof2D(double x1[],double x2[],double y[],double sig[],int ndata,
	      double a[],int ia[],
	int ma, double **alpha, double beta[], double *chisq,
	void (*funcs)(double, double, double [], double *, double [],int))
#else
void mrqcof2D(x1,x2, y, sig, ndata, a, ia, ma,  alpha,  beta,  chisq, funcs)  
  double x1[],x2[],y[],sig[],a[],**alpha,beta[],*chisq;
  int ndata,ma,ia[];
  void (*funcs)();
#endif
{
	int i,j,k,l,m,mfit=0;
	double ymod,wt,sig2i,dy,*dyda;

	dyda=dvector(1,ma);
	for (j=1;j<=ma;j++)
		if (ia[j]) mfit++;
	for (j=1;j<=mfit;j++) {
		for (k=1;k<=j;k++) alpha[j][k]=0.0;
		beta[j]=0.0;
	}
	*chisq=0.0;
	for (i=1;i<=ndata;i++) {
		(*funcs)(x1[i],x2[i],a,&ymod,dyda,ma);
		sig2i=1.0/(sig[i]*sig[i]);
		dy=y[i]-ymod;
		for (j=0,l=1;l<=ma;l++) {
			if (ia[l]) {
				wt=dyda[l]*sig2i;
				for (j++,k=0,m=1;m<=l;m++)
					if (ia[m]) alpha[j][++k] += wt*dyda[m];
				beta[j] += dy*wt;
			}
		}
		*chisq += dy*dy*sig2i;
	}
	for (j=2;j<=mfit;j++)
		for (k=1;k<j;k++) alpha[k][j]=alpha[j][k];
	free_dvector(dyda,1,ma);
}
