/*
  polarimetry.c    by Thomas Szeifert, Landessternwarte Heidelberg    21.10.97

  
  tools to get the Stokes parameters from the normalized differences between
  ordinary and extraordinary rays after passing retarder plates and Wollaston
  prism.
*/




#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <proto_fors.h>

#ifndef myabs
#define myabs(A) ((A)  >  0.0 ? (A) : (-A))
#endif
#define RAD1     (180.0 / acos(-1.0))
#ifndef PI
#define PI       acos(-1.0)
#endif
/*---------------------------------------------------------------------------*/

#ifdef __STDC__
void sfourier(float ao[], float ao_err[], float ae[], float ae_err[], 
	      float phi,float Q[] ,float Qerr[], float U[], float Uerr[], 
	      int irow, int ipos)
#else
void sfourier (ao,ao_err, ae,ae_err,phi,Q,Qerr,U,Uerr,irow,ipos)
  float ao[],ae[];
  float phi,ao_err[],ae_err[];
  float Q[],Qerr[],U[],Uerr[];
  int irow,ipos;
#endif
{
  
  float buffer;
  int row;

  /* 

     Compute Stokes parameter -- add to the Stokes U and Q components the 
     normalized flux differences at one of the ipos discrete retarder plate 
     angles. 
     
     Q = Q + 2/ipos * (ao-ae)/(ao+ae) cos(4*phi)
     U = U + 2/ipos * (ao-ae)/(ao+ae) sin(4*phi)

     */

  for (row = 0; row < irow; row++)
    {
      if   (ao[row] - ae[row] != 0){
	Q[row] = Q[row] + (2.0/ipos)*((ao[row]-ae[row])/(ao[row]+ae[row])) * 
	  (float) cos((double) (4.0 * phi/RAD1));
      }
      if   (ao[row] - ae[row] != 0){
	U[row] = U[row] + (2.0/ipos)*((ao[row]-ae[row])/(ao[row]+ae[row])) * 
	  (float) sin((double) (4.0 * phi/RAD1));
      }
      

      /* calculation of the mean square deviation from mean value:*/
      /* valid for small ao[row]_err/ao[row] and ae_err[row]/ae[row] */
      /* relative error of the quotient (ao[row]-ae[row])/(ao[row]+ae[row]): */

      if   (ao[row]-ae[row] != 0.0 && ao[row]+ae[row] != 0.0 ){
	buffer = (float) sqrt((DSQR(ao_err[row])+DSQR(ae_err[row]))/DSQR(ao[row]-ae[row])+(DSQR(ao_err[row])+DSQR(ae_err[row]))/DSQR(ao[row]+ae[row]));
      }  
      else{
	buffer = 1.0e+31; 
      }

      /* sum square of absolute error: */ 
      Qerr[row] = Qerr[row] + SQR((2.0/ipos)*((ao[row]-ae[row])/(ao[row]+ae[row]))*(float) sin( 4*phi/RAD1)*buffer);
      /* cos not sin TS 1999-03-14 */
      Uerr[row] = Uerr[row] + SQR((2.0/ipos)*((ao[row]-ae[row])/(ao[row]+ae[row]))*(float) cos( 4*phi/RAD1)*buffer);

    }


}









#ifdef __STDC__
int stokesi (float *ima, float *err, float *buffima, 
	    float *buffsig, int npix, int num, int mode)
#else
int stokesi (ima, err, buffima, buffsig, npix, num, mode)
     float *ima,*err,*buffima,*buffsig;
     int npix, num, mode;
#endif

#define INIT -1
#define SUM 0
#define AVERAGE 1
{

  /* Compute Stokes I, intensity == (sum over all frames)/(number of frames) */

  int i;
  float sig2;

  switch (mode)
    {
    case SUM:
      for (i=0; i<npix; i++)/*sum for all retarder angles*/ 
	{
	  ima[i] = ima[i]+buffima[i];
	  sig2 = buffsig[i]*buffsig[i];
	  err[i] = err[i]+sig2;
	}
      break;
    case AVERAGE:/*(mean I) = (sum) / (number of retarder positions) */
      for (i=0; i<npix; i++)
	{
	  ima[i] = 2.0*ima[i]/((float) num);  /* average over all expos*/
	  err[i] = (float) (sqrt(err[i]/(double) num) / 
			    sqrt((double)(num-1)));/*num independent*/
	}
      break;
    case INIT:/*set sums to zero*/
      for (i=0; i<npix; i++)
	{
	  err[i] = 0.0;
	  ima[i] = 0.0;
	}
    }
  return 0;
}








/*
  compute degree of polarization p and angle of p[olarization theta from
  the Stokesparameters Q and U

  the offset caused by the non-linear transformation of Q,U to p,theta
  for finit standard deviations is also considered if imode is set to 1
  */
#ifdef __STDC__
int linpol(float Q[], float Qerr[], float U[], float Uerr[],
       float *pol, float *polerr, float *theta, float *thetaerr,
       int npix, int nstep, int imode)
#else
int linpol (Q, Qerr, U, Uerr, pol, polerr, theta, thetaerr, npix,nstep, imode)
     float Q[], Qerr[], U[], Uerr[];
     float *pol, *polerr, *theta, *thetaerr;
     int npix,nstep,imode;
#endif

{
  register int i;  
  register float  degree2,lsqrsum;


  register int iQ,iU;
  register float gauss,sump,sump2,sumt,sumt2,sumgauss,Qp,Up,buff,theta0,QU2n;

  /*  Compute linear polarizationm angle and corresponding errors
      from Stokes Q, U and corresponding errors.
      Algorithm see linpol.for in spol/src (Jeremy Walsh) 
      */


  for (i=0; i < npix; i++)
    {
      sump = sump2 = sumt = sumt2 = sumgauss = 0.0;
 
/* calculate angle and degree of polarization and the mean deviation 
   with non-linear statistics assuming gaussian distribution for Q and U
   which will be ok if Q and U are both small*/

      if ((Q[i] == 0) && (U[i] == 0)) 
	{
	  pol[i] = (float) 0.0;
	  polerr[i] = (float) sqrt (DSQR(Qerr[i]) + DSQR(Uerr[i])); 
	  theta[i] = (float) 45.0;
	  thetaerr[i] = (float) 51.96154; 
	  /*
	    51.96154 == 12**(-0.5) * PI * RAD 
	    said Serkovski (Advances in Astron & Astroph. 1, 1962)
	    but it's not really applicable because only true for polarisation
	    expressed in magnitudes!
	    */	  
	}
      else if (imode == 0)
	{
	  /* if imode == 0 we don't correct for statistical bias */
	  pol[i] = (float) sqrt(DSQR(Q[i]) + DSQR(U[i]));
	  theta[i] = (float) RAD1 * 0.50 * (float) atan2(U[i],Q[i]);
	  theta[i] = theta[i] > (float) 0. ?  theta[i] : 
	    ((float) 180.0) + theta[i];
	}
      else
	{/* scalar product of function and gaussian distribution */	  
	  if (Uerr[i] > 1.0e-10 && Qerr[i] > 1.0e-10){
	    
	    QU2n = 0.5/(PI * SQR(((float) (nstep))/5.0) );
	    theta0 = RAD1*0.50 * (float) atan2((double) U[i],(double) Q[i]);

	    for (iQ = 0 ; iQ <= 2*nstep ; iQ++)
	      {/* here we integrate over 5 sigma in Q*/
		Qp = Qerr[i]*( (float) (iQ - nstep)/(((float) nstep)/5.))+Q[i];
		for (iU = 0 ; iU <= 2*nstep ; iU++)
		  {/* here we integrate over 5 sigma in U*/
		    Up = Uerr[i]*((float)(iU-nstep)/((float) nstep/5.)) + U[i];
		    gauss = QU2n * (float) exp(-0.5*(DSQR((Up-U[i])/Uerr[i]) +
				DSQR( (Qp-Q[i])/Qerr[i] )));
		    sumgauss = sumgauss + gauss;

		    lsqrsum = SQR(Up) + SQR(Qp);
		    sump = sump + (float) (sqrt(lsqrsum)) * gauss;
		    sump2 = sump2 + lsqrsum * gauss;
		
		    buff = RAD1*0.50*(float) atan2((double) Up, (double) Qp);
		    if (theta0 < 0){ /* fix priblems close to 0 or 180 degree*/
		      buff = (fabs(buff-theta0) < fabs(buff-180.0-theta0)) ? 
			buff : buff - 180.0;
		    }
		    else{
		      buff = (fabs(buff-theta0) < fabs(buff+180.0-theta0)) ? 
			buff : buff + 180.0;
		    }
		    sumt = sumt + buff * gauss;
		    sumt2 = sumt2 + SQR(buff) * gauss;
		  }
	      }


	    /*for security reason sums are renormalized: 
	      mostly: (sumgauss - 1 ~< 1.0e-6)*/
	    sumt = sumt/sumgauss;
	    sumt2 = sumt2/sumgauss;
	    sump = sump/sumgauss;
	    sump2 = sump2/sumgauss;


	    
	    /* sigma = sqrt( <P(U,Q)**2> - <P(U,Q)>**2) */  
	    polerr[i] = (float) sqrt( sump2 - DSQR(sump));
	    thetaerr[i] = (float) sqrt(sumt2 - DSQR(sumt));


	    /* this is: sqrt(SQR(Q[i])+SQR(U[i])) minus the difference between
	       Erwartungsvalue and sqrt(SQR(Q[i])+SQR(U[i])) */
	    /* if p ~= 1 the integral in the Q and U space exceeds 1 - however
	       this is unlikely in reality and the correction has to be done
	       also for Q and U in subroutine sfourier and later for p */  
	    pol[i] = 2.0 * (float) sqrt(DSQR(Q[i])+DSQR(U[i])) - sump;

	    buff = (float) (RAD1 * 0.50 * atan2((double) U[i], (double) Q[i]));
     	    theta[i] = 2.0 * buff - sumt;

	    theta[i] =  ( theta[i] < 0.0 ) ? (theta[i] + 180.0) :  theta[i];

	  }
	  else{ /* if sigma = Qerr = 0 or Uerr = 0 the gaussian make trouble */
	    polerr[i] = (float) sqrt (DSQR( Q[i]*Qerr[i]) + DSQR( U[i]*Uerr[i]))/
	      (DSQR(Q[i]) + DSQR(U[i]));
	    degree2 = (SQR(Q[i]) + SQR(U[i]) - SQR(polerr[i]));
	    if (degree2 > 0.0) 
	      {
		pol[i] = (float) sqrt (degree2);
	      }
	    else
	      {
		pol[i] = (float) 0.0;
	      }
	    theta[i] = (float)(RAD1 * 0.50 * atan2 ((double) U[i], (double) Q[i]));
	    theta[i] = ( theta[i] < 0 ) ? theta[i] + 180.0 :  theta[i];
	    thetaerr[i] = (float) (RAD1 * 0.50 * sqrt(DSQR(Uerr[i]*Q[i]) + 
	      DSQR(Qerr[i]*U[i]))/(DSQR(U[i])+DSQR(Q[i])));
	  }
       }


    }

  return 0;
}









/*
  ---------------------------------------------------------------------------
  
  decomissioned  method with quotient of two retarder positions 
  at phi and phi + 45 degree for linear polarimetry or 
  at phi = 45 and phi = -45 degree for circular polarimetry

  */

#ifdef __STDC__
int stokes (float ao[], float ae[], float bo[], float be[],
       float ao_err[], float ae_err[], float bo_err[], float be_err[],
       float ima[], float err[], int npix)
#else
int stokes (ao,ae,bo,be,ao_err,ae_err,bo_err,be_err,ima,err,npix)
     float ao[],ae[],bo[],be[];
     float ao_err[],ae_err[],bo_err[],be_err[];
     float ima[],err[];
     int npix;
#endif
{
  
  /*  Compute Stokes parameter by the quotient method.
      Two spectra from two frames are needed. Retarder 
      angle should be 0 and 45 or 22.5 and 67.5 degrees for linear
      or -45 an 45 degrees for circular polarization.
      Formula see source code of Jeremy Walsh, stokes.for

      Sabine & Otmar Version - (not longer used by software)
      */
 
  float a,b,r,v,absb,absa;
  float sigv,sigaoda,sigaeda,sigbodb,sigbedb;
  int i;

  for (i=0; i <npix; i++)
    {
      a = (ae[i] <= 0.0) ? 1.0e+30 : ao[i]/ae[i];
      b = (be[i] <= 0.0) ? 1.0e+30 : bo[i]/be[i];
      absa = myabs(a);
      absb = myabs(b);
      r = (absb <= 0.0) ? 1.0e+30 : sqrt(absa/absb);
      v = (absb <= 0.0) ? 1.0 : (r-1.0)/(r+1.0);
      if (ae[i] <= 0) 
	v = 1; 
      else if (bo[i] == 0)
	v = 1;
      ima[i] = v;
      
      sigaoda = (ao[i] <= 0.0) ? 0.0 : ao_err[i]/ao[i]; 
      sigaoda = sigaoda*sigaoda;
      sigaeda = (ae[i] <= 0.0) ? 0.0 : ae_err[i]/ae[i]; 
      sigaeda = sigaeda*sigaeda;

      sigbodb = (bo[i] <= 0.0) ? 0.0 : bo_err[i]/bo[i]; 
      sigbodb = sigbodb*sigbodb;
      sigbedb = (be[i] <= 0.0) ? 0.0 : be_err[i]/be[i]; 
      sigbedb = sigbedb*sigbedb;

      printf(" i %6d  r %10.5f \n",i,r);
      sigv = sqrt(sigaoda + sigaeda + sigbodb + sigbedb) * r / (SQR(r) +1.0);
      err[i] = (float) sigv;
    }
  return 0;
}









#ifdef __STDC__
int stokesv (float ao[], float ae[], float bo[], float be[],
       float ao_err[], float ae_err[], float bo_err[], float be_err[],
       float ima[], float err[], int npix)
#else
int stokesv (ao,ae,bo,be,ao_err,ae_err,bo_err,be_err,ima,err,npix)
     float ao[],ae[],bo[],be[];
     float ao_err[],ae_err[],bo_err[],be_err[];
     float ima[],err[];
     int npix;
#endif
{
  
  /* 
     calculate circular polarimetry == Stokes V

     Retarder angle must be 315 = -45 and 45 degrees for circular polarization.
     
     */
 
  int i;

  for (i=0; i <npix; i++) /* for any table entry */
    {
      ima[i] = (float) 0.5 * ((ao[i] - ae[i])/(ao[i] + ae[i]) - /*-45 degree*/ 
			      (bo[i] - be[i])/(bo[i] + be[i])); /* 45 degree*/

      if ( (ao[i]-ae[i] != 0) && (ao[i]+ae[i] != 0) &&
	   (bo[i]-be[i] != 0) && (bo[i]+be[i] != 0)){
	err[i] =  /* 45 degree*/
	  SQR((float) 0.5 * (ao[i] - ae[i])/(ao[i] + ae[i])) * ( 
	  (SQR(ao_err[i]) + SQR(ae_err[i])) / SQR(ao[i]-ae[i]) + /*Zaehler*/
	  (SQR(ao_err[i]) + SQR(ae_err[i])) / SQR(ao[i]+ae[i])   /*Nenner*/
	  )
	  +       /*-45 degree*/
	  SQR((float) 0.5 * (bo[i] - be[i])/(bo[i] + be[i])) * ( 
	  (SQR(bo_err[i]) + SQR(be_err[i])) / SQR(bo[i]-be[i]) + /*Zaehler*/
	  (SQR(bo_err[i]) + SQR(be_err[i])) / SQR(bo[i]+be[i])   /*Nenner*/
	  );
	err[i] = (float) sqrt (err[i]);
      }
      else{
	err[i] = (float) 1.0;
      }
    }
  return 0;
}








