/*lib_ipolmagni.c                    by Thomas Szeifert LSW Heidelberg   10/96 

This program maps data for a given part of the frame (a slitlet of a 
polarimetric mask) and computes:
   -> sky
   -> point spread function
   -> flux
for all stars which were selected in the input table. We search for stars that
are in a given slitlet defined by a second input table and loop through this
table. 

We add the pixel value in the innermost ring    r < inputi[0] 
and add the fit value within the second ring    r < inputi[1] 
and add the fit value for all pixels outside the mask and for multiple stars

From my point of view the single Moffat profile (PSF=1 or PSF=2) fits well 
enough for polarimetry, were we don't need absolute fluxes. If you are going 
to use the program for absolute photometry you will have to change the point 
spread function with set/ipol PSF=3

|  PSF=3 means:
|  take a sum of two Moffat functions with different exponents beta1 & beta2
|  ->  funcs = f * ( (1 - epsilon**2)/ arg**beta1 + epsilon**2 / arg**beta2)
|  beta1 = 2; beta2 = 7 
|  see function moffat for more explanations. 

see R. Racine 1996 - PASP 108, 699 for PSF fitting and more advanced models
*/



/* WARNING:  extern arrays, variables and pointers inside proto_ipolmagni.h:*/
#include <math.h>
#include <midas_def.h>

#include <proto_fors.h>
#include <proto_ipolmagni.h>



#define MAXSLITS 100
#define MAXSTARS 1000
#ifndef PI 
#define PI acos(-1.0)
#endif



#ifdef __STDC__
  int reject(float x,float y,float lxpxl[],float lypxl[],int msel ,int j,
	     int param)
#else
     int reject(x,y,lxpxl,lypxl,j,msel,param)
     int j,msel,param;
     /*    float x,y,xpxl[],ypxl[];*/ 
     float x,y,lxpxl[],lypxl[];
#endif
{
  int i,rej;
  float r;

  rej = 0;
  for (i = 0 ; i < msel ; i++) 
    {
      r= (float) sqrt( SQR(lxpxl[i]-x) + SQR(lypxl[i]-y) );
      if (r < ((float) param) && i != j)
	rej = 1;
    }

  return rej;
}




#ifdef __STDC__
  float fraction(int ix, int iy, float x, float y, int irad)
#else
  float fraction(ix,iy,x,y,irad)
    int ix,iy,irad;
    float x,y;
#endif
{
  float frac,buff;

  buff = (float) sqrt(SQR(x-ix)+SQR(y-iy));

  if (f2i(buff) < irad)
    frac = 1.0;
  else if (f2i(buff) == irad)
    frac = 0.5 - (buff - (float) irad);
  else
    frac = 0.0;

  return frac;
}


/**********************************************************************
  The very famous Moffat function!     T.Szeifert LSW-Heidelberg  10/96
  ____
  \                          Fi                 
Y= \  -------------------------------------------------------------- 
   /  (1+((x-xi)/Ax)**2+((y-yi)/Ay)**2-2*Axy*(x-xi)(y-yi)/(AxAy)**2)**beta
  /___
 i=1,msel

***********************************************************************/

/*
  here I assume that the sky is subtracted in the higher level routines!
  the coordinates must be correct! If you want to calculate coordinates
  with this routine you have to change higher level routines and you 
  have to calculate the first deviates dY/dxi and dY/dyi for any i<=msel

  x1,x2 are the input coordinates - in pixels
  dyda[1..msel+4] are the first partial deviates  d_*y / d_a[i]
  of *y and a[1..msel+4] are the parameters
  xpxl ypxl are the coordinates of the selected stars (msel stars are selected)
  a[1] is the exponent -- beta
  a[2],a[3] are the PSF's widths -- Ax,Ay
  a[4] corrects for inclined eliptical profiles -- Axy
  a[5..msel] are the peek intensities of the PSF -- Fi
  the other variable are needed to avoid repetetive calculations - mostly

*/

#ifdef __STDC__
  void moffat(double x1,double x2, double a[], double *y, double dyda[], 
	      int inull)
#else
  void moffat( x1,x2, a, y, dyda, inull)
  double x1,x2,a[],*y,dyda[];
  int inull;
#endif
{
  double adyda,arg,xmx,ymy,xmx2,ymy2,xmy,twoa1,ma1,twoa1a2,twoa1a3,m2a4;
  int i;
  
  *y = 0.0;
  twoa1 = a[1]*2.0;
  twoa1a2 = twoa1/a[2];
  twoa1a3 = twoa1/a[3];
  m2a4 = (-2.0)*a[4];
  ma1 = a[1]*(-1.0);
	

  for (i = 1 ; i <=4 ; i++)
    dyda[i] = 0.0;

  for (i = 1; i <= ncoef-4; i++)
    {
      if (a[2] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[2]);
      xmx = (x1 - xsel[i])/a[2];
      xmx2 = SQR(xmx);
      if (a[3] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[3]);
      ymy = (x2 - ysel[i])/a[3];
      ymy2 = SQR(ymy);
      xmy = m2a4*xmx*ymy;

      arg = 1.0 + xmx2 + ymy2 + xmy; 
      if (arg <= 0) printf("fatal Moffat error -> logarithm of %11.5 ",arg);
      dyda[i+4] = pow(arg,ma1);
      adyda = a[i+4]*pow(arg,(ma1-1));

      *y = *y + a[i+4]*dyda[i+4];
	
      dyda[1] = dyda[1] - a[i+4]*log(arg)*dyda[i+4];
      dyda[2] = dyda[2] + twoa1a2*(xmx2+xmy)*adyda;
      dyda[3] = dyda[3] + twoa1a3*(ymy2+xmy)*adyda;
      dyda[4] = dyda[4] +  twoa1*xmx*ymy*adyda;
    }


}


#ifdef __STDC__
  void moffatbf(double x1,double x2, double a[], double *y, double dyda[],
		int inull)
#else
  void moffatbf( x1,x2, a, y, dyda, inull)
  double x1,x2,a[],*y,dyda[];
  int inull;
#endif
{
  double adyda,arg,xmx,ymy,xmx2,ymy2,xmy,twoa1,ma1,twoa1a2,twoa1a3,m2a4;
  int i;
  
  *y = 0.0;
  twoa1 = dbeta[0]*2.0;
  twoa1a2 = twoa1/a[1];
  twoa1a3 = twoa1/a[2];
  m2a4 = (-2.0)*a[3];
  ma1 = dbeta[0]*(-1.0);
	

  for (i = 1 ; i <=3 ; i++)
    dyda[i] = 0.0;

  for (i = 1; i <= ncoef-3; i++)
    {
      if (a[1] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[1]);
      xmx = (x1 - xsel[i])/a[1];
      xmx2 = SQR(xmx);
      if (a[2] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[2]);
      ymy = (x2 - ysel[i])/a[2];
      ymy2 = SQR(ymy);
      xmy = m2a4*xmx*ymy;

      arg = 1.0 + xmx2 + ymy2 + xmy; 
      if (arg <= 0) printf("fatal Moffat error -> logarithm of %11.5 ",arg);
      dyda[i+3] = pow(arg,ma1);
      adyda = a[i+3]*pow(arg,(ma1-1));

      *y = *y + a[i+3]*dyda[i+3];
	

      dyda[1] = dyda[1] + twoa1a2*(xmx2+xmy)*adyda;
      dyda[2] = dyda[2] + twoa1a3*(ymy2+xmy)*adyda;
      dyda[3] = dyda[3] +  twoa1*xmx*ymy*adyda;
    }
}



#ifdef __STDC__
  void moffatbf2(double x1,double x2, double a[], double *y, double dyda[],
		 int inull)
#else
  void moffatbf2( x1,x2, a, y, dyda, inull)
  double x1,x2,a[],*y,dyda[];
  int inull;
#endif
{
  double adyda[2],arg,xmx,ymy,xmx2,ymy2,xmy,twoa1[2],ma1[2],twoa1a2[2],
    twoa1a3[2],m2a4;
  int i,j;
  
  *y = 0.0;

  for (j = 0 ; j < 2 ; j++)
    {
      twoa1[j] = dbeta[j]*2.0;
      twoa1a2[j] = twoa1[j]/a[1];
      twoa1a3[j] = twoa1[j]/a[2];
      ma1[j] = dbeta[j]*(-1.0);
    }

  m2a4 = (-2.0)*a[3];

	

  for (i = 1 ; i <=4 ; i++)
    dyda[i] = 0.0;

  for (i = 1; i <= ncoef-4; i++)
    {
      if (a[1] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[1]);
      xmx = (x1 - xsel[i])/a[1];
      xmx2 = SQR(xmx);
      if (a[2] == 0) printf("fatal Moffat error -> devision by %11.5 ",a[2]);
      ymy = (x2 - ysel[i])/a[2];
      ymy2 = SQR(ymy);
      xmy = m2a4*xmx*ymy;

      arg = 1.0 + xmx2 + ymy2 + xmy; 

      dyda[i+4] = pow(arg,ma1[0]) * ( 1.0 - DSQR(a[4])) + pow(arg,ma1[1]) 
	* DSQR(a[4]);

      *y = *y + a[i+4]*dyda[i+4];
	
      adyda[0] = (1.0-DSQR(a[4])) * a[i+4]*pow(arg,(ma1[0]-1.0));
      adyda[1] = DSQR(a[4])       * a[i+4]*pow(arg,(ma1[1]-1.0));
      

      dyda[1] = dyda[1] + twoa1a2[0]*(xmx2+xmy)*adyda[0] +
	                  twoa1a2[1]*(xmx2+xmy)*adyda[1];
      dyda[2] = dyda[2] + twoa1a3[0]*(ymy2+xmy)*adyda[0] +
	                  twoa1a3[1]*(ymy2+xmy)*adyda[1];
      dyda[3] = dyda[3] + twoa1[0]*xmx*ymy*adyda[0] + 
	                  twoa1[1]*xmx*ymy*adyda[1];
      dyda[4] = dyda[4] + a[i+4] * (pow(arg,ma1[0]) * (-2.0*a[4]) 
				    + 2.0 * a[4] * pow(arg,ma1[1]));
    }
}



/************************************************************
 *
 * fit2D -- fitting of Moffat function 
 * 
 * calls   : mrqmin 
 * modified: Criterium of stopping is more relaxed (C.Levin).
 * modified: application for PSF-fitting with Moffat functions (T.Szeifert)
 *
 ************************************************************/
#define EPS	0.0003      /* delta chi**2 / chi**2    TS-0896*/
#ifdef __STDC__
   void fit2D( double *x1, double *x2, double *y, int n, 
	       double *a )
#else
   void fit2D( x1, x2, y, n, a )
   double *x1,*x2,*y,*a;
   int n;
#endif
{
    int *ia;
    int nfit = ncoef, ncoefs = ncoef;
    int i, iter = 1, j;
    register double **covar, **alpha;
    double chisq, ochisq, alamda = -1.;
    register double *sig;
    double sigsum=0,sqrsum=0;


    /* initialize guess parameters of the fit: */
    if (reset == 1){
      if (psf == 1){
	a[1] = 4.0;                             /* canonic value of DAOPHOT */
	for (j = 2; j <= 3; j++) a[j] = 6.0;    /* must be greater than HWHM */
	a[4] = 0.0;                             /* no elongation */
	reset = 0;
      }
      else if (psf == 2){
	for (j = 1; j <= 2; j++) a[j] = 6.0;    /* must be greater than HWHM */
	a[3] = 0.0;                             /* no elongation */
	reset = 0;
      }
      else if (psf == 3){
	for (j = 1; j <= 2; j++) a[j] = 6.0;    /* must be greater than HWHM */
	a[3] = 0.0;                             /* no elongation */
	a[4] = 0.0;
	reset = 0;
      }
    }
      
    if (guess != 1)
      for (j = fcoef+1; j <= ncoef; j++) a[j] = 65536.0; 
    /* start with 2**16 as central intensity because of better nmerical
       convergence for decreasing fluxes !!! Don't change it !!! If you
       do so, you'll get faster results for most -- but a crash for some
       bright objects. (I presume 16 bit CCD readout)*/

    sig = dvector( 1, n );
    ia = ivector( 1, nfit );
    covar = dmatrix( 1, nfit, 1, nfit );
    alpha = dmatrix( 1, nfit, 1, nfit );


    for ( i = 1; i <= n; i++ ){

      
      sig[i] = sqrt(fabs(y[i] + skybuff + rparam[0]/rparam[1])/rparam[1]);

      
      sigsum = sigsum + SQR(1.0/sig[i]);
      sqrsum = sqrsum + SQR(y[i]);
      
      /* equal weights would change a little:*//*
      sig[i] = 1.0;
      sigsum = sqrsum = 1.0;
      */
    }


    for ( i = 1; i <= nfit; i++ )
	ia[i] = 1;



    if (debug >= 2){
      printf("   ...fit2D: guess values\n");
      for (j = 1; j <= ncoef; j++)
	printf(" %15.7f",a[j]);
      printf("\n");
    }


    mrqmin2D( x1, x2, y, sig, n, a, ia, ncoefs, covar, alpha,
		&chisq, evalpsf, &alamda );
    if (debug >= 1)
      printf("   ...fit2D: %2d iterations %7d pixels ==> chi = %9.7f\n", iter, n, sqrt(chisq));

    if (debug >= 2){
      for (j = 1; j <= ncoef; j++)
	printf(" %15.7f",a[j]);
      printf("\n");
    }

    do {
	iter++;
	ochisq = chisq;

	mrqmin2D( x1, x2, y, sig, n, a, ia, ncoefs, covar, alpha,
		  &chisq, evalpsf, &alamda );


	if (debug >= 1){
	  sprintf(text,"             %2d iterations %7d pixels ==> chi = %9.7f", iter, n, sqrt(chisq));
	  SCTDIS(text,80);

	}
	if (debug >= 2){
	  for (j = 1; j <= ncoef; j++)
	    printf(" %15.7f",a[j]);
	  printf("\n");
	}
    } while ( fabs(ochisq - chisq) / chisq > EPS || iter <= 6) ; 

    if (debug == 0)
      printf("\n   ...fit2D: %2d iterations %7d pixels ==> chi = %9.7f\n",
	     iter, n, sqrt(chisq));

    if (debug >= 1)
      printf("\n              wchi =%8.3f   wsum =%8.1f   wchi/wsum =%8.5f\n", sqrt(chisq), sqrt(sqrsum/sigsum), sqrt(chisq/(sqrsum/sigsum)));


    alamda = 0.;  /* To de-allocate memory */

    mrqmin2D( x1, x2, y, sig, n, a, ia, ncoefs, covar, alpha,
		&chisq, evalpsf, &alamda );


    /*
      if (sqrt(chisq/sqrsum) >= 3 * noise)
      printf("         WARNING: high residuals >= 3 sigma\n");
      */

    free_dvector( sig, 1, n );
    free_ivector( ia, 1, nfit );
    free_dmatrix( covar, 1, nfit, 1, nfit );
    free_dmatrix( alpha, 1, nfit, 1, nfit );
}




#ifdef __STDC__
  void evalpsf(double x1, double x2, double a[], double *y, double dyda[], int itmp)
#else
  void evalpsf( x1, x2, a, y, dyda, itmp)
  double x1,x2,a[],*y,dyda[];
  int itmp;
#endif
{
  int inull = 0;

  if (psf == 1)
    moffat( x1, x2, a, y, dyda, inull);
  else if (psf == 2)
    moffatbf( x1, x2, a, y, dyda, inull);
  else if (psf == 3)
    moffatbf2( x1, x2, a, y, dyda, inull);
  else
    printf("ERROR: undefined function \n");

  return;
}
/*game over*/





