/*ipolmagni.c                        by Thomas Szeifert LSW Heidelberg   10/96 

This program maps data for a given part of the frame (a slitlet of a 
polarimetric mask) and computes:
   -> sky
   -> point spread function
   -> flux
for all stars which were selected in the input table. We search for stars that
are in a given slitlet defined by a second input table and loop through this
table. 

We add the pixel value in the innermost ring    r < inputi[0] 
and add the fit value within the second ring    r < inputi[1] 
and add the fit value for all pixels outside the mask and for multiple stars

From my point of view the single Moffat profile fits well enough for 
polarimetry, were we don't need absolute fluxes. If you are going to use
the program for absolute photometry you will have to change the program 
in the following way:

|  take a sum of two Moffat functions with different exponents beta1 & beta2
|  ->  funcs = f * ( (1 - epsilon)/ arg**beta1 + epsilon / arg**beta2)
|  do some analysis to get the first deviates analytically and
|  see function moffat for more explanations. You have to change the fitparam
|  array and ncoef (fcoef = 6 but not 4) because of the two more coeficients!

see R. Racine 1996 - PASP 108, 699 for PSF fitting and more advanced models
*/




/*__________________________________________________________________________*/

/* WARNING:  these arrays, variables and pointers are given as extern 
             parameters to the functions via proto_ipolmagni.h*/

double Start[2], Step[2], dbeta[2], skybuff;
float rparam[3];
int Npix[2];
int debug, reset, guess; 
/* debug can be 0,1 or 2; higher == more informations*/
int ncoef, psf, fcoef=0; 
/* Moffat needs fcoef = 4 fit parameters for the PSF if beta is free*/
/* fitting routines math functios moffat... need the coordinates via extern*/
float *xsel,*ysel,*fsel;
char text[80];

/* WARNING:  these were the arrays cited above */
/*__________________________________________________________________________*/



/* WARNING:  declaration of extern arrays, variables and pointers is inside 
             proto_ipolmagni.h:*/
#include <proto_ipolmagni.h>

#include <math.h>
#include <tbldef.h>
#include <midas_def.h>
#include <proto_fors.h>

#define MAXSLITS 100
#define MAXSTARS 1000
#ifndef PI
#define PI (float) acos(-1.0)
#endif
#ifndef RAD
#define RAD (float) 180.0/PI
#endif

#ifdef __STDC__
void mapslit (int,float,float,int,int,float []);
void compusky(float [],float,float,int,float[],float[],int,float[],float[],
	      int[], float *, float *);
double fitmoffat(float [], float, float, int,  float [],	float [], int, 
	       float [], float [], int [],float [], double [], float *, 
		 float *);
void compuflux(float [],float,float,int,float[],float[],int,float[],float[],
	       int[],float[],float[],double [],double, float *, float *);
#else
void mapslit ();       /*read the data from the disk*/
void compusky ();      /*median the sky region*/
double fitmoffat();      /*fit the point spread function*/
void compuflux ();     /*integrate flux in the ring*/
#endif

int main ()
{
  int msel;
  int i, j, actval;
  int kun, knul;
  float ystart[MAXSLITS], yend[MAXSLITS];/*in_mask*/
  int ocol[10];
  /*in_mask: - prefix n*/
  int ncol,nrow,nsort,allcol,allrow,incol[3],nsel,innr[MAXSLITS],inull,nid=-1;
  /*in_star: - prefix m*/
  int mcol,mrow,msort,imcol[3],tid,*imnr;
  int iselect,ymaxrow;
  int fid,iunit;
  int istart=0,ipxl;
  int iparam[4];

  char qualif[5], in_image[61], in_mask[61],in_star[61],out_table[61];
  char cnlog[MAXSLITS];
  char instruma[72], original[61], unit[64], line[80];

  /* float tab[5], offset[MAXSLITS], low, rbff;*/
  float *image;
  float *xcen,*ycen,*fcen;
  float *sky,*skyerr,*flux,*fluxerr;
  float mag,magerr;

  double sumpsf;     /*integral of normalized PSF*/  
  double *fitparam,dparam[4];
  double buff;


  float *xpxl,*ypxl;       


  /* get into Midas and get parameters */

  SCSPRO ("ipolmagni");
  SCTPUT ("\nMAGNI/IPOL");
  strcpy(instruma,""); 
  strcpy(unit,"");


  SCKRDI  ("DEBUG", 1, 1, &actval, &debug, &kun, &knul);
  /* read the keywords */
  if (debug >= 1)
    SCTPUT ("\n...read the keywords");

  SCKGETC ("QUALIF",  1, 4, &actval, qualif);
  SCKGETC ("ORI",   1, 60, &actval, original);
  SCKGETC ("IN_A",   1, 60, &actval, in_image);
  SCKGETC ("IN_B",   1, 60, &actval, in_mask);
  SCKGETC ("IN_C",   1, 60, &actval, in_star);
  SCKGETC ("OUT_A",  1, 60, &actval, out_table);
  SCKRDI  ("PSF", 1, 1, &actval, &psf, &kun, &knul);  
  SCKRDI  ("INPUTI", 1, 4, &actval, iparam, &kun, &knul);  /*apertures*/
  SCKRDR  ("INPUTR", 1, 3, &actval, rparam, &kun, &knul);  /*ccd-param*/
  SCKRDD  ("INPUTD", 1, 4, &actval, dparam, &kun, &knul); /*fit-param*/


  if (debug >= 1) {
    printf ("      IMAGE:      %s          coordinates from table   %s \n",in_image,in_star);
    printf ("      APERTURES:  deny fit %d    fit %d    inner sky %d    outer sky %d\n",iparam[0],iparam[1],iparam[2],iparam[3]);
    if (psf == 1) 
      printf ("      FUNCTION:   a la Moffat    beta free\n");
    if (psf == 2) 
      printf ("      FUNCTION:   a la Moffat    beta = %5.2f\n",dparam[0]);
    if (psf == 3) 
      printf ("      FUNCTION:   a la Moffat    (1-eps**2) * F(beta=2) + eps**2 * F(beta=7)\n");
    printf ("      CCD:        ron = %4.1f e    gain = %4.1f e/adu    clip = %4.1f sigma\n",rparam[0],rparam[1],rparam[2]);
  }

  if (debug >= 1){
    if (iparam[0] == 0 )
      printf("   ...INFO: fit all pixels - iparam[0] = 0 \n");
    if (iparam[0] >= iparam[1] )
      printf("   ...INFO: deny any fit if possible - iparam[0] >= iparam[1] \n");
    if (iparam[2] >= iparam[3] || iparam[2] <= 1 )
      printf("   ...INFO: skip compusky - iparam[2] >=  iparam[3] or iparam[2] < 1 \n", iparam[2],iparam[3]);
  }

  /* open sesam */
  if (debug >= 1)
    SCTPUT ("...open sesam");

  SCFOPN(in_image, D_R4_FORMAT, 0, F_IMA_TYPE, &fid);
  SCDRDD(fid,"START",1,2,&actval,Start,&iunit,&inull);
  SCDRDD(fid,"STEP",1,2,&actval,Step,&iunit,&inull);
  SCDRDI(fid,"NPIX",1,2,&actval,Npix,&iunit,&inull);
  

  if (in_mask[0] != '-') 
    {


      /* read-select the input table with the mask*/
      if (debug >= 1)
	SCTPUT ("...read-select the input table with the mask");

      TCTOPN(in_mask, F_IO_MODE, &nid);
      TCIGET (nid, &ncol, &nrow, &nsort, &allcol, &allrow);
      TCCSER(nid,":ystart",&incol[0]);
      TCCSER(nid,":yend",&incol[1]);
      TCCSER(nid,":ray_typ",&incol[2]);
      if (incol[0] == -1 || incol[1] == -1 || incol[2] == -1)
	{
	  SCTPUT ("missing columns in input table with mask");
	  SCTPUT ("one of the columns :ystart,:yend,:ray_type doesn't exist");
	  SCSEPI ();
	  exit (1);
	}
      ymaxrow= 0;
      nsel = 0;
      for (j = 0 ; j < nrow ; j++)
	{
	  TCSGET(nid,j+1,&iselect);
	  if (iselect == 1)
	    {
	      TCERDR(nid,j+1,incol[0],&ystart[nsel],&inull);
	      TCERDR(nid,j+1,incol[1],&yend[nsel],&inull);
	      if (ystart[nsel]-yend[nsel]+1 > ymaxrow) 
		ymaxrow = ystart[nsel]-yend[nsel]+1;
	      TCERDC(nid,j+1,incol[2],&cnlog[nsel],&inull);
	      innr[nsel] = j; 
	      nsel++;
	    }
	}
    }
  else
    {
      ystart[0] = (float) Start[1];
      yend[0] = (float) (Start[1] + Npix[1] * Step[1]);
      nsel = 1;
    }


  /* read-select the input table with the target stars and the neighbors */
  if (debug >= 1)
    SCTPUT ("...read-select the input table with the target stars");

  TCTOPN(in_star, F_IO_MODE, &tid);
  TCSCNT(tid,&msel);

  xcen = vector(0,msel-1);
  ycen = vector(0,msel-1);
  fcen = vector(0,msel-1);
  imnr = ivector(0,msel-1);

  TCIGET (tid, &mcol, &mrow, &msort, &allcol, &allrow);
  TCCSER(tid,":XCEN",&imcol[0]);
  TCCSER(tid,":YCEN",&imcol[1]);
  TCCSER(tid,":ICENT",&imcol[2]);
  if (imcol[0] == -1 || imcol[1] == -1)
    {
      SCTPUT ("missing columns in input table with star");
      SCTPUT ("one of the columns :XCEN,:YCEN doesn't exist");
      SCSEPI ();
      exit (1);
    }
  guess = 0;
  msel = 0;
  for (j = 0 ; j < mrow ; j++)
    {
      TCSGET(tid,j+1,&iselect);
      if (iselect == 1)
	{
	  TCERDR(tid,j+1,imcol[0],&xcen[msel],&inull);
	  TCERDR(tid,j+1,imcol[1],&ycen[msel],&inull);
	  if (imcol[2] > -1){
	    TCERDR(tid,j+1,imcol[2],&fcen[msel],&inull);
	    guess = 1;
	  }
	  imnr[msel] = j; 
	  msel++;
	}
    }


  xpxl = vector(0,msel-1);
  ypxl = vector(0,msel-1);
  sky = vector(0,msel-1);
  skyerr = vector(0,msel-1);
  flux = vector(0,msel-1);
  fluxerr = vector(0,msel-1);

  xsel = vector(1,msel);
  ysel = vector(1,msel);
  fsel = vector(1,msel);


  /*****  Here we set up the function parameters:   *****/

  /* ignore the guess if somethings wrong with the input*/
  if (dparam[0] <= 0)
    reset = 1; 
  /* number of parameters for the normalized profile    */
  if (psf == 1)
    fcoef = 4;
  else if (psf == 2){
    fcoef = 3;
    dbeta[0] = dparam[0];
    printf("dbeta = %6.2f  dparam[0] = %6.2f \n",dbeta[0],dparam[0]);
  }
  else if (psf == 3){
    fcoef = 4;
    dbeta[0] = 7.0;
    dbeta[1] = 2.0;
  }

  /*  fitparam = (double *) osmmget((msel+fcoef+1) * sizeof(double));*/
  fitparam = dvector(0,msel+fcoef+1);

  /* start parameters for better convergence */
  if (reset != 1){
    if (psf == 1){
      fitparam[0] = 3.0;
      for (i = 1 ; i < fcoef ; i++)
	fitparam[i] = dparam[i];
    }
    else if (psf == 2){
      for (i = 0 ; i < fcoef ; i++)
	fitparam[i] = dparam[i+1]/(2.0*sqrt(( pow(2.0,(1.0/dbeta[0])) - 1.0)));
    }
    else if (psf == 3){
      for (i = 0 ; i < fcoef-1 ; i++)
	fitparam[i] = dparam[i+1];
      fitparam[0] = fitparam[0] * 1.5498;
      fitparam[1] = fitparam[1] * 1.5498;
      fitparam[fcoef-1] = 0.1;
    }
  }

  /*************************  ***************************/



  /* loop the slitlets */
  for (j = 0 ; j < nsel; j++)
    {

      ipxl = Npix[0]*( f2i((float) ((yend[j] - ystart[j])/Step[1]) ) + 1);
      image = vector(0,ipxl-1);

      if (debug >= 1)
	SCTPUT ("\n...mapslit");/* map slit */
      mapslit(fid, ystart[j], yend[j],istart, ipxl, image);
      /*in:  ystart,yend - mark the slit position 
	out: istart,ipxl - first pixel and number of pixel to read
	     image       - array with the data*/
      /*      for (i = 0 ; i <= 10 ; i++) printf("%10.2f",image[i]);*/

      if (debug >= 1)
	SCTPUT ("...compusky");/* compute sky - median in cleaned ring */
      compusky(image, ystart[j], yend[j],ipxl,xcen,ycen,msel,sky,skyerr,
	       iparam, xpxl, ypxl);
      /*in:  like mapslit
	     image          - array with input data 
             xcen,ycen,msel - position of the sources, number of sources
	     iparam         - integration circle, no mans land, sky circle
	out: sky,skyerr     - the output sky values and errors*/


      if (debug >= 1)
	SCTPUT ("...fitmoffat");/* fit PSF by Moffat function */
      sumpsf = fitmoffat(image,ystart[j], yend[j],ipxl,xcen,ycen,msel,sky,
			 skyerr,iparam,fcen,fitparam, xpxl, ypxl);
      /*in:  like mapslit & compusky
	out: fitparam       - numerical parameters of the PSF first the
	                      the 4 parameters of the Moffat function: beta
			      Ax,Ay,Axy and second the central intensities fi
			      of the star's profiles
             sumpsf         - integral of PSF with peek intensity f = 1.0*/


      if (debug >= 1)
	SCTPUT ("...compuflux");/* integrate flux in a circle */
      compuflux(image, ystart[j], yend[j],ipxl,xcen,ycen,msel,sky,skyerr,
		iparam,flux,fluxerr,fitparam,sumpsf, xpxl, ypxl);
      /*in:  like mapslit & compusky & fitmoffat
	out: flux, fluxerr  - the flux of the sources and the errors*/

      free_vector(image,0,ipxl-1);
    }

  SCFCLO (fid);



  free_vector(xcen,0,msel-1);
  free_vector(ycen,0,msel-1);
  free_vector(xpxl,0,msel-1);
  free_vector(ypxl,0,msel-1);


  

  if (debug >= 1)
    SCTPUT ("\n...write output into table");
  switch (toupper(qualif[0]))
    {
    case ('I') :
      {
	TCCSER (tid, ":sky", &ocol[0]);
	if (ocol[0] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F10.2", "unitless", "sky", &ocol[0]);
	TCCSER (tid, ":skyerr", &ocol[1]);
	if (ocol[1] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F10.2", "unitless", "skyerr", 
		  &ocol[1]);
	TCCSER (tid, ":flux", &ocol[2]);
	if (ocol[2] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F10.2", "unitless", "flux", 
		  &ocol[2]);
	TCCSER (tid, ":fluxerr", &ocol[3]);
	if (ocol[3] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F10.2", "unitless", "fluxerr", 
		  &ocol[3]);
	TCCSER (tid, ":fit", &ocol[6]);
	if (ocol[6] == -1)
	  TCCINI (tid, D_R8_FORMAT, 1, "F10.2", "unitless", "fit", 
		  &ocol[6]);
	TCCSER (tid, ":mag", &ocol[4]);
	if (ocol[4] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F7.3", "magnutides", "mag", 
		  &ocol[4]);
	TCCSER (tid, ":magerr", &ocol[5]);
	if (ocol[5] == -1)
	  TCCINI (tid, D_R4_FORMAT, 1, "F7.3", "magnutides", "magerr", 
		  &ocol[5]);
	TCCSER (tid, ":magfit", &ocol[7]);
	if (ocol[7] == -1)
	  TCCINI (tid, D_R8_FORMAT, 1, "F7.3", "unitless", "magfit", 
		  &ocol[7]);

	for (i = 0 ; i < msel ; i++)
	  {
	    TCEWRR(tid,imnr[i]+1,ocol[0],&sky[i]);
	    TCEWRR(tid,imnr[i]+1,ocol[1],&skyerr[i]);
	    TCEWRR(tid,imnr[i]+1,ocol[2],&flux[i]);
	    TCEWRR(tid,imnr[i]+1,ocol[3],&fluxerr[i]);

	    mag = (flux[i] > 0.0) ?  (float)((-2.5)*log10(flux[i])) : 99.99;
	    magerr = (flux[i] > 0.0) ? (fluxerr[i]/flux[i]) : 99.99;
	    if (mag < 99.98){
	      TCEWRR(tid,imnr[i]+1,ocol[4],&mag);
	      TCEWRR(tid,imnr[i]+1,ocol[5],&magerr);
	    }
	    TCEWRD(tid,imnr[i]+1,ocol[6],&fitparam[i+fcoef]);
	    if (fitparam[i+fcoef] > 0.0){
	      buff = log10(fitparam[i+fcoef]) * (-2.5);
	      TCEWRD(tid,imnr[i]+1,ocol[7],&buff);
	    }
	  }
	/* Additional column ray_typ (O(rdinary) or E(xtra-ordinary)) */
	/*	 TCCINI (tid, D_C_FORMAT, 1, "A1", "", "RAY_TYP", &icol[4]);*/
	break;
      }
    default:
      {
        sprintf(line,"Error: Unknown qualifier %s\n",qualif);
        SCTPUT(line);
        SCETER(9, "Exiting...");
        break;
      }
    }

/*          epilogue ...                 */

  TCTCLO (tid);
  if (nid != -1)
    TCTCLO (nid);



  free_vector(sky,0,msel-1);
  free_vector(skyerr,0,msel-1);

  free_vector(flux,0,msel-1);
  free_vector(fluxerr,0,msel-1);
  free_ivector(imnr,0,msel-1);
  free_dvector(fitparam,0,msel+fcoef+1);

  free_vector(xsel,1,msel);
  free_vector(ysel,1,msel);
  free_vector(fsel,1,msel);



  SCTPUT ("");
  SCSEPI ();
  exit (0);
}
/*end of main*/




#ifdef __STDC__
  void mapslit(int fid, float yst, float yen, int istart, int ipxl, 
	       float image[]) 
#else
  void mapslit(fid,yst,yen,istart,ipxl,image)
    int fid, istart, ipxl;
    float image[], yst, yen;
#endif
{
  int actsize;
  
  istart = Npix[0] * f2i( (float) ((yst-Start[1])/Step[1]) ) + 1;
  SCFGET(fid, istart, ipxl, &actsize, (char *) image); 
  if (debug >= 2)
    printf("   ...1st pxl = %6d  Npxl = %6d  1st pxl value =%8.1f\n", istart,ipxl,image[0]);

  return;
}




#ifdef __STDC__
  void compusky(float image[], float yst, float yen, int ipxl,  float xcen[],
		float ycen[], int msel, float sky[], float skyerr[], 
		int iparam[], float *xpxl,float *ypxl) 
#else
  void compusky(image,yst,yen,ipxl,xcen,ycen,msel,sky,skyerr,iparam,xpxl,ypxl)
    int ipxl,msel,iparam[];
    float image[], yst,  yen, xcen[], ycen[], sky[], skyerr[],*xpxl,*ypxl;
#endif
{
  int i,j,imulti[MAXSTARS],pxl,ix,iy,npxl,rej,jsel;
  register float r,*buff,*mbuff,sumsig,noise;


  buff = vector(1,SQR((iparam[3]*2+1)));
  mbuff = vector(1,SQR((iparam[3]*2+1)));



  for (j=0;j<msel;j++)
    {
      xpxl[j] = (float) ((xcen[j]-Start[0])/Step[0] + 1.0 );
      ypxl[j] = (float) ((ycen[j]-yst)/Step[1] + 1.0 );
    }


  if (iparam[3] > iparam[2] && iparam[2] > 1){
    for (i=0;i<msel;i++)
      {
	if ( ycen[i] > yst && ycen[i] < yen){
	  /* check multiplicity -- sky rings may overlap*/ 
	  imulti[i]=0;
	  imulti[i] = reject(xpxl[i],ypxl[i],xpxl,ypxl,msel,i,iparam[2]*2);
	  /* pik pixels in the ring then pik median of these pxl's-values */
	  /*loop square...*/
	  npxl = 0;
	  for (iy = f2i(ypxl[i])-iparam[3]-1; iy <=  f2i(ypxl[i])+iparam[3]+1;
	       iy++){
	    for (ix = f2i(xpxl[i])-iparam[3]-1; ix <= f2i(xpxl[i])+iparam[3]+1;
		 ix++){
	      if (ix >=0 && ix < Npix[0]){  /*check if inside slit - x*/
		pxl = (iy - 1) * Npix[0] + ix - 1; /*TS*/
		if (pxl>=0 && pxl<ipxl){    /*check if inside slit - y*/
		  r = (float) sqrt((double) SQR((float)ix - xpxl[i]) +
				   (double) SQR((float)iy - ypxl[i]));
		  /*check if inside ring:*/
		  if (f2i(r) <= iparam[3] && f2i(r) >= iparam[2]){
		    if (imulti[i]==0)
		      {
			buff[++npxl] = image[pxl];/*starts at 1--nrutils!*/
			mbuff[npxl] = buff[npxl];
		      } 
		    else
		      {                         /*check if ix,iy is clean*/
			/*ix,iy may almost touch skyrings of other sources:*/
			rej = reject((float)ix,(float)iy,xpxl,ypxl,msel,i,
				     iparam[2]);
			if (rej==0){ 
			  buff[++npxl] = image[pxl];
			  mbuff[npxl] = buff[npxl];
			}
		      }
		  }
		}
	      }
	    }
	  }
	  /*...end loop*/
	  
	  jsel = 0;
	  sky[i] =  pik_median(npxl,mbuff);/*warning pik_median sorts mbuff!*/
	  noise = (float) rparam[2] * (float)
	    sqrt(SQR(rparam[0]/rparam[1]) + sky[i]/rparam[1]);
	  for (j = 1;j<=npxl;j++)
	    {/*reject bad pixels, cosmix & stars if >= rparam[2] times noise*/
	      if ((float) (fabs(buff[j]-sky[i])) < noise) {
		mbuff[++jsel] = buff[j]; 
		buff[jsel] = buff[j];
	      }
	    }
	  sky[i] = pik_median(jsel,mbuff);
	  /*warning pik_median sorts mbuff!*/

	  sumsig = 0.0;
	  for (j = 1;j<=jsel;j++)
	    sumsig = sumsig + SQR(buff[j]-sky[i]);
 	  skyerr[i] = (float) sqrt(sumsig/(jsel*jsel-1));
	  skybuff = sky[i]; /*buffer for fit2D*/
	  if (debug >= 2)
	    printf("star = %6d sky = %10.2f skyerr = %10.2f \n",i+1,sky[i],skyerr[i]);
	}/*if ycen*/
      }/*for i*/
  }
  else{
    for (i=0;i<msel;i++)
      {  
	sky[i] = 0.0;
	skyerr[i] = 0.0;
      }
  }


  free_vector(buff,1,SQR((iparam[3]*2+1)));
  free_vector(mbuff,1,SQR((iparam[3]*2+1)));

  return;
}





/*---------------------------------------------------------------------------*/
#ifdef __STDC__
  double fitmoffat(float image[], float yst, float yen, int ipxl, float xcen[],
		   float ycen[], int msel, float sky[], float skyerr[], 
		   int iparam[], float fcen[], double fitparam[], 
		   float *xpxl,float *ypxl) 
#else
  double fitmoffat(image,yst,yen,ipxl,xcen,ycen,msel,sky,skyerr,iparam,
		   fcen, fitparam,xpxl,ypxl)
    int ipxl,msel,iparam[];
    float image[], yst,  yen, xcen[], ycen[], fcen[], sky[], skyerr[],
    *xpxl, *ypxl;
    double fitparam[];
#endif
{
  int i,pxl,ix,iy,npxl,*isel,ibuff,inull=0;
  float r;
  double *x,*y,*f,*buff,*buff1,*nulldyda;
  double sumff,ff,frac,piRsqr,phi;

  /*  point = 0;*/
  npxl = 0;
  ncoef = fcoef;

  isel = ivector(0,msel);
  buff = dvector(1,msel+fcoef+2);
  buff1 = dvector(1,msel+fcoef+2);
  nulldyda = dvector(1,fcoef+2);
  x = dvector (1,msel*SQR(2*iparam[1]+1));
  y = dvector (1,msel*SQR(2*iparam[1]+1));
  f = dvector (1,msel*SQR(2*iparam[1]+1));

  for (i=0;i<msel;i++)
    {
      if ( ycen[i] > yst && ycen[i] < yen){
	ncoef++;
	xsel[ncoef-fcoef] = xpxl[i];
	ysel[ncoef-fcoef] = ypxl[i];
	if (guess == 1)
	  fsel[ncoef-fcoef] = fcen[i];
	isel[ncoef-(fcoef+1)] = i;
	
	/*loop square...*/	  /* pik pixels in the ring */
	for (ix = f2i(xpxl[i])-iparam[1]-1;ix <= f2i(xpxl[i])+iparam[1]+1;ix++)
	  {
	    if (ix >= 0 && ix < Npix[0]){   /*check if inside slit - x*/
	      for (iy  = f2i(ypxl[i])-iparam[1]-1; 
		   iy <= f2i(ypxl[i])+iparam[1]+1; iy++)
		{
		  pxl = (iy -1) * Npix[0] + ix - 1; /*TS*/
		  if (pxl>=0 && pxl<ipxl){   /*check if inside slit - y*/
		    r = (float) sqrt(SQR(ix - xpxl[i])+SQR(iy-ypxl[i]));
		    if (f2i(r) < iparam[1]+1){
		      x[++npxl] = ix;
		      y[npxl] = iy;
		      f[npxl] = image[pxl]-sky[i];
		    }
		  }
		}
	    }
	  }
      }/*if ycen*/
    }/*for i*/

  if (ncoef > 43){
    printf("WARNING: you have selected %4d stars - that's a lot! stop program and select the input table",(ncoef-3));
  }

  if (ncoef > fcoef) /*this means there are selected stars*/ 
    {
      for (i = 0; i < fcoef; i++) buff[i+1] = fitparam[i];/*set start values*/
      if (guess == 1){
	for (i = fcoef+1; i <= ncoef; i++) buff[i] = fsel[i-fcoef];
      }
      fit2D (x,y,f,npxl+1,buff); /*fit the PSF for msel stars*/
      for (i = 0; i < fcoef ; i++) fitparam[i] = buff[i+1];
      for (i = 1; i <= fcoef+1 ; i++) buff1[i] = buff[i];



      /*calculate integral of the normalized (maximum peak=1) PSF*/
      ibuff = ncoef;
      ncoef = fcoef+1;
      buff1[ncoef] = 1.0;
      sumff = 0.0;
      piRsqr = 0.0;
      for (ix = f2i(xsel[1])-iparam[1]-1; ix <= f2i(xsel[1])+iparam[1]+1; ix++)
	{ 
	  for (iy  = (f2i(ysel[1])-iparam[1]-1); 
	       iy <= (f2i(ysel[1])+iparam[1]+1); iy++)
	    {
	      r = (float) sqrt(SQR(ix - xsel[1])+SQR(iy-ysel[1]));
	      if (f2i(r) < iparam[1]+1){
		evalpsf((double) ix, (double) iy,buff1,&ff,nulldyda,inull);
		frac = fraction(ix,iy,xsel[1], ysel[1],iparam[1]);
		piRsqr = piRsqr + frac;
		sumff = sumff + ff * frac;

		pxl = (iy - 1) * Npix[0] + ix - 1; /*TS*/

		if (ix >= (xsel[1]-1) && ix <= (xsel[1]) && debug == 2 && pxl > 0 && pxl < ipxl)
		  printf ("      %4d %4d pxl= %4d fit= %6.0f flux= %6.0f residual= %6.0f\n",ix,iy,pxl,(buff[fcoef+1]*ff),(image[pxl]-sky[0]),(buff[fcoef+1]*ff)-(image[pxl]-sky[0]));

		if ( (iy - f2i(ysel[1])) >= -4 && (iy - f2i(ysel[1])) < 4 && debug == 3) {
		  printf ("%5.0f%5.0f",(buff[fcoef+1]*ff),(image[pxl]-sky[0]));
		}
	      }
	    }
	  if (debug >= 2)
	    printf ("\n");
	}
      ncoef = ibuff;
      for (i = 0; i < ncoef-fcoef;i++) 
	fitparam[isel[i]+fcoef] = sumff*buff[i+fcoef+1];


      printf("   ...parameters:");
      if (psf == 1){	
	printf("\n      beta  = %9.6f   Ax  = %9.6f   Ay  = %9.6f   Axy  = %9.6f",fitparam[0],fitparam[1],fitparam[2],fitparam[3]);
	phi = pow(fitparam[1],2.0) -  pow(fitparam[2],2.0) != 0.0 ? 
	  atan2(2.0*fitparam[3]*fitparam[2]*fitparam[1],(pow(fitparam[1],2.0) -  pow(fitparam[2],2.0))) : 0.0;
	printf("\n                          FWHM(x) = %5.2f   FWHM(y) = %5.2f   phi  = %9.1f",2.0*fitparam[1]*sqrt(pow(2.0,1.0/fitparam[0])-1.0),2.0*fitparam[2]*sqrt(pow(2.0,1.0/fitparam[0])-1.0),RAD*phi*0.5);
      }
      else if (psf == 2){
	printf("\n      Ax  = %9.6f   Ay  = %9.6f   Axy  = %9.6f",fitparam[0],fitparam[1],fitparam[2]);
	phi = pow(fitparam[1],2.0) -  pow(fitparam[2],2.0) != 0.0 ? 
	  atan2(2.0*fitparam[2]*fitparam[1]*fitparam[0],(pow(fitparam[0],2.0) -  pow(fitparam[1],2.0))) : 0.0;
	printf("\n      FWHM(x) = %5.2f   FWHM(y) = %5.2f   phi  = %9.1f",2.0*fitparam[0]*sqrt(pow(2.0,1.0/dbeta[0])-1.0), 2.0*fitparam[1]*sqrt(pow(2.0,1.0/dbeta[0])-1.0),RAD*phi*0.5);
      }
      else if (psf == 3){
	printf("\n      Ax  = %9.6f   Ay  = %9.6f   Axy  = %9.6f   eps**2 = %9.6f ",fitparam[0],fitparam[1],fitparam[2],DSQR(fitparam[3]));
      }
      if (debug >= 0){
	printf("\n   ...integrated fit:\n");
	for (i = fcoef; i < msel+fcoef ;i++) 
	  printf (" %5d %12.4f ",(i-fcoef+1),fitparam[i]);
	printf("\n   ...peek flux:\n");
	for (i = fcoef; i < msel+fcoef ;i++) printf (" %5d %12.4f ",i-fcoef+1,fitparam[i]/sumff);


	printf ("\n   ...integral of normalized profile = %13.6f on %11.6f pixels\n",sumff,piRsqr);
      }
      if ( fabs( PI * SQR(iparam[1]) - piRsqr)/piRsqr > 0.005){
 	printf("      WARNING: integration of PSF exceeds error limit (0.005)\n");
	printf("      due to small integration radius:    A = PI*r**2 = %11.6f pixels\n",(PI * SQR(iparam[1])));
      }

    }      


  free_ivector(isel,0,msel);
  free_dvector(buff,1,msel+fcoef+2);
  free_dvector(buff1,1,msel+fcoef+2);
  free_dvector(nulldyda,1,fcoef+2);
  free_dvector(x,1,msel*SQR(2*iparam[1]+1));
  free_dvector(y,1,msel*SQR(2*iparam[1]+1));
  free_dvector(f,1,msel*SQR(2*iparam[1]+1));

  return sumff;

}




/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void compuflux(float image[], float yst, float yen, int ipxl,  float xcen[],
	       float ycen[], int msel, float sky[], float skyerr[], 
	       int iparam[],float flux[],float fluxerr[], double fitparam[],
	       double sumpsf,float *xpxl,float *ypxl) 
#else
     void compuflux(image,yst,yen,ipxl,xcen,ycen,msel,sky,skyerr,iparam,
		    flux,fluxerr,fitparam,sumpsf,xpxl,ypxl)
     float image[], yst, yen, xcen[], ycen[], sky[], skyerr[], flux[], fluxerr[], *xpxl,*ypxl;
     double sumpsf, fitparam[]; 
     int ipxl, msel, iparam[];
#endif
{

  float r,sumflux,sumpxl,sumerr,frac;
  double *dbuff,ff,*nulldyda;
  int i,ix,iy,pxl,inull=0,imulti[MAXSTARS];

  dbuff  = dvector (1 , fcoef+msel+1);
  nulldyda = dvector (1 , fcoef+msel+1);

  for (i=1;i<=fcoef;i++) 
    {
      dbuff[i] = fitparam[i-1]; /*profile parameters*/
      nulldyda[i] = 0; /*buffer for the first partial deviates*/
    }
  ncoef = fcoef+1; /* to evaluate the PSF */

  for (i=0;i<msel;i++)
    {/*for i*/
      if ( ycen[i] > yst && ycen[i] < yen) {
	sumflux = 0.0;
	sumpxl = 0.0;
	sumerr = 0.0;
	imulti[i]=0;
	xsel[1] = xpxl[i];
	ysel[1] = ypxl[i];
	dbuff[fcoef+1] = fitparam[i+fcoef] / sumpsf;
	
	/* check multiplicity -- integr circle may touch other skyrings*/ 
	imulti[i] = reject(xpxl[i],ypxl[i],xpxl,ypxl,msel,i,
			     iparam[1]+iparam[2]);

	/* pik pixels in the ring */
	if (imulti[i]==0 && iparam[0] > 0){
	  /*loop square...*/
	  for (ix = f2i(xpxl[i])-iparam[1]-1; ix <= f2i(xpxl[i])+iparam[1]+1;
	       ix++)
	    {/*for ix*/
	      if (ix >= 0 && ix < Npix[0]){      /*check if inside slit - x*/
		for (iy = f2i(ypxl[i])-iparam[1]-1; 
		     iy <= f2i(ypxl[i])+iparam[1]+1; iy++)
		  {/*for iy*/
		    pxl = (iy - 1) * Npix[0] + ix - 1;/*TS*/
		    if (pxl>=0 && pxl<ipxl){     /*check if inside slit - y*/
		      r= (float)sqrt(SQR(ix - xpxl[i])+SQR(iy-ypxl[i]));
		      if (f2i(r) < iparam[1]+1){     /*if pixel inside ring*/
			frac = fraction(ix,iy,xpxl[i],ypxl[i],iparam[1]);
			if (iparam[0] >= iparam[1]){     /*deny any fitting*/
			  sumflux = sumflux + (image[pxl]-sky[i]) * frac;
			  sumpxl = sumpxl + frac; 
			  sumerr = sumerr + (image[pxl]*frac)/
			    (float) sqrt(rparam[1]) + SQR(frac*skyerr[i]);
			}
			else if (f2i(r) <= iparam[0]){
			                       /*deny fit for r < iparam[0]*/
			  sumflux = sumflux + (image[pxl]-sky[i]) * frac;
			  sumpxl = sumpxl + frac; 
			  sumerr = sumerr + (image[pxl]*frac)/
			    (float) sqrt(rparam[1]) + SQR(frac*skyerr[i]);
			}
			else{         /* fit if  iparam[0] < r < iparam[1] */
			  evalpsf((double)ix,(double)iy, dbuff, &ff, nulldyda,
				  inull); 
			  sumflux = sumflux + (float) ff * frac;
			  sumpxl = sumpxl + frac; 
			  sumerr = sumerr + ((float) ff*frac)/
			    (float) sqrt(rparam[1]) + SQR(frac*skyerr[i]);
			}
		      }
		    }
		    else { /* else moffat - it's outside slit in x*/
		      r= (float) sqrt(SQR(ix - xpxl[i])+SQR(iy-ypxl[i]));
		      if (f2i(r) < iparam[1]+1){
			frac = fraction(ix,iy,xpxl[i],ypxl[i],iparam[1]);
			evalpsf((double) ix,(double) iy, dbuff, &ff, nulldyda,
				inull); 
			sumflux = sumflux + (float) ff * frac;
			sumpxl = sumpxl + frac; 
			sumerr = sumerr + ((float)ff*frac)/
			  (float) sqrt(rparam[1]) + SQR(frac*skyerr[i]);
		      }		      
		    }
		  }
	      }	      
	      else { /* else moffat - it's outside slit in y*/
		r= (float) sqrt(SQR(ix - xpxl[i])+SQR(iy-ypxl[i]));
		if (f2i(r) < iparam[1]+1){
		  frac = fraction(ix,iy,xpxl[i],ypxl[i],iparam[1]);
		  evalpsf((double) ix,(double) iy, dbuff, &ff, nulldyda, 
			  inull); 
		  sumflux = sumflux + (float) ff * frac;
		  sumpxl = sumpxl + frac; 
		  sumerr = sumerr + ((float) ff*frac)/
		    (float)sqrt(rparam[1]) + SQR(frac*skyerr[i]);
		}		      
	      }

	    }
	  flux[i] = (float) sumflux;
	  fluxerr[i] = (float) sqrt(sumerr);
	} /* ...end if multi */
	else {        /* multiple star - no aperture photometrie is possible */
	  /* therefore we take the fitted magnitudes */
	  flux[i] = (float) fitparam[i+fcoef]; 
	  fluxerr[i] = (float) (sqrt(flux[i]/sqrt(rparam[1]) + 
			 SQR(skyerr[i]*PI*SQR(iparam[1]))));
	}
	/* else moffat */
      }/*if ycen*/
    }/*for i*/
  

  printf("   ...integrated flux:\n");
  for (i = 0; i < msel ;i++) 
    printf (" %5d %12.4f ",(i+1),flux[i]);

  if (debug <= 0) printf("\n");

  /*
  printf("fcoef\n");
  for (i = 1 ; i <= fcoef+msel+1 ; i++)
    printf("%f %f %f\n",dbuff[i],nulldyda[i],fitparam[i-1]);
    */

  /*
  free_dvector( dbuff, 1, fcoef+msel+1);
  free_dvector( nulldyda, 1, fcoef+msel+1);
  */

  return;
}

/*game over*/






