/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* .IDENT       mosdisp2D.c                                */
/* .AUTHORS     Pascal Ballester (ESO/Garching)    LONG    */
/*              Cristian Levin   (ESO/La Silla)    LONG    */
/*              Sabine Moehler   (LSW/Heidelberg)  MOS     */
/*              Thomas Szeifert  (LSW/Heidelberg)  MOS_2D  */
/* .KEYWORDS    Spectroscopy, Long-Slit, MOS               */
/* .PURPOSE                                                */
/* .VERSION     March 1996                                 */
/* ------------------------------------------------------- */
/* Functions (in this module):

   spectroscopic context:

   setrefdeg_2D      setdisp_2D    mos_fit_disp_2D   mos_eval_disp_2D
   mos_writedisp_2D  mos_savedisp  printdisp         setrefdeg
   set_zero          setdisp       mos_fit_disp      mos_eval_disp
   mos_initdisp      finishdisp    mos_readdisp      mos_writedisp

   numerical context:


   fleg_2D   fcheb_2D   fpoly_2D
 */


/* Purpose:
   This module is related to the dispersion relation and its representation.
   The upper modules (moscalib.c, mosrebin.c) do not know which representation
   is used (Legendre polynomials, or any other), nor how it is stored (which
   descriptor, meaning of coefficients). This design enables to replace 
   this module mosdisp.c by any other based on a different representation.
*/

/* Note:
`
   Because the module is based on Numerical Recipes routines, which arrays
   start at index 1, the general rule is that arrays start at index 1. However
   the module can accomodate any starting index, which is specified at 
   initialisation time (function initdisp).
*/


#include <math.h>
#include <midas_def.h>
#include <tbldef.h>

#include <proto_fors.h>
#include <proto_mos.h>

#ifndef PI
#define PI acos(-1)
#endif

#ifndef SQR
static float sqrarg;
#define SQR(a) (sqrarg=(a),sqrarg*sqrarg)
#endif

#define MAXNCOE 100



int FIT_SUCCESS = 0; /* Sucess indicator for the fit */





/*-------------------------------------------------------------------------

  WARNING:   Here we define some automatic variables that are frequently
             used below in the dispersion routines!	     
	     These variables occur without any reference in the functions. 
	     They are hidden in main() although any function of the module
	     comunicates with the others of the module. A change of degy by 
	     setrefdeg_2D would be noticed by fleg_2D during the next call 
	     of the function!
*/

double   pixbin;                        /* Pixel bin in wavelength units */
double   coef[100];                            /* Array of coefficients  */
int	 fdeg, ncoef, refdeg, maxcoef; /* Number of coefficients, Degree */
int      degy,degxy; /* order of polynoms in y-dir and of the mixed term */
int	 tide, colslit, colline, coly, colrms, colcoef[100];
int	 start_index, nbline;

/*
  WARNING:   These were the automatic variables

-------------------------------------------------------------------------*/








/* ------------------------------------------------------------------------
Description:

   The module provides ten top-level routines related to the evaluation of the
   dispersion relation. Only those six routines can be used in applications 
   related to the dispersion relation (wavelength calibration, rebinning, ...).
   The lower level routines must be considered as private.

   --- void mos_savedisp (save) ---

   Stores dispersion coefficients in array <save>.

   --- double mos_fit_disp (ndata, degree, x, l, chi) ---
   --- double mos_fit_disp_2D (ndata, degree, x, l, chi) ---

   This routine evaluates the dispersion relation on the basis of ndata 
   couples of d values (x,l). Array indexes start at 0. The degree is 
   indicative. If ndata is too small, a smaller degree will be fitted 
   to the couples. Minimum value for ndata is 2. The procedure returns 
   a rough estimate of the average pixel size and the error of the 
   dispersion relation.

   --- void mos_eval_disp (x, l, ndata) ---
   --- void mos_eval_disp_2D (x, l, ndata) ---

   Applies the last previously estimated (or loaded) dispersion relation 
   to ndata values of x and estimates the corresponding wavelength l. 
   Arrays start at index 0.


   --- void mos_initdisp (name, mode, start) ---

   Open (mode = OLD) or Create (mode = NEW) a storage table (named <name>).
   In addition to initdisp also the column :SLIT is created.

   --- void finishdisp() ---

   Closes the storage table

   --- void mos_writedisp(line, slit, pix_y, world_y, nyrow) ---
   --- void mos_writedisp_2D(line, slit, pix_y, world_y, nyrow) ---

   Store in the storage table the dispersion relation corresponding to the 
   slitlet <slit>, row number <line>, and world coordinate <y>.

   --- double mos_readdisp(line) ---

   Retrieves from the storage table the dispersion relation corresponding 
   to the row number <line>.  The procedure returns a rough estimate of 
   the average pixel size.

   --- void setdisp(degree, coefs) ---
   --- void setdisp_2D(degree, coefs) ---

   Sets the coefficients of the dispersion relation. coefs is an array
   which element 0 is the constant term and other elements represent
   the coefficients of higher degree

---------------------------------------------------------------------------- */








/*  ------------------------------------------------------------------------

    Here the 2D functions by Thomas Szeifert

    ------------------------------------------------------------------------ */


/* ------------------------------------------------------------------------- */
#ifdef _STDC__
  void setrefdeg_2D(int deg, int deg1, int deg2)
#else
  void setrefdeg_2D(deg, deg1, deg2) 
  int deg,deg1,deg2;
#endif
{

refdeg = deg;
degy = deg1;
degxy = deg2;
maxcoef = refdeg + degy + degxy + 1;
}

/* ------------------------------------------------------------------------- */
#ifdef _STDC__
  void setdisp_2D(int deg, double coefs[])
#else
  void setdisp_2D(deg, coefs)
  int deg;
  double  coefs[];
#endif
{

int i;

fdeg = deg;
refdeg = deg;
ncoef = deg + degy + degxy + 1;
maxcoef = deg + degy + degxy + 1;

for (i=1; i<=ncoef; i++) 
  coef[i] = coefs[i-1];
}


/* -------------------------------------------------------------------------*/


#ifdef _STDC__
  double mos_fit_disp_2D(int *ndata, int *deg, double x[], double y[],
		       double l[], double *chi)
#else
  double mos_fit_disp_2D(ndata, deg, x, y, l, chi)
  int *ndata, *deg;
  double x[],y[],l[],*chi;
#endif
{

  double   *sig, **covar, chisq;
  int *lista;
  int i;
  int maxdeg,actvals;
  char poltyp[8];

  /* printf("NData : %d   ---   Deg value : %d\n",*ndata,*deg); */

  refdeg = *deg;
  maxdeg = *ndata ;
  if (refdeg > maxdeg) fdeg = maxdeg;
  else                 fdeg = refdeg;
  ncoef   = fdeg + degy + degxy + 1;
  maxcoef = fdeg + degy + degxy + 1;

  if (*ndata < ncoef) {
    printf("Not enough lines (minimum is 2). \nNo dispersion relation computed\n");
    return(-2.);
  }
  
  if (fdeg < 1)  {
    printf("Degree : %d. No dispersion relation fitted\n",*deg);
    return(-1.);
  }
  

  covar=dmatrix(1,*ndata,1,*ndata);
  /*  chisq=(double*)dvector(0,*ndata);*/
  sig  =(double*)dvector(1,*ndata);
  
  lista=ivector(1,ncoef);
  
  for (i=1; i<=ncoef; i++) lista[i] = 1;
  for (i=1; i<=*ndata; i++) sig[i]=1.;

  SCKGETC("POLTYP", 1, 8, &actvals, poltyp);
  if (toupper(poltyp[0]) == 'L')
    dlfit_2D(x,y,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,fleg_2D);
  else if (toupper(poltyp[0]) == 'C')
    dlfit_2D(x,y,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,fcheb_2D);
  else if (toupper(poltyp[0]) == 'P')
    dlfit_2D(x,y,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,fpoly_2D);
  else 
    printf("ERROR - You have specified an invalid polynom type \n"); 

  *chi = chisq;
  
  free_dmatrix(covar,1,*ndata,1,*ndata);
  /*  free_dvector(chisq,0,*ndata);*/
  free_dvector(sig,1,*ndata);
  free_ivector(lista,1,ncoef);
  
  pixbin = coef[2];
  return(pixbin);
  
}


/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void mos_eval_disp_2D (double x[], double y[], double l[], int n)
#else
void mos_eval_disp_2D (x, y, l, n)
     double x[], y[], l[];
     int n;
#endif

{

  int actvals, i, icoef;
  double xp[MAXNCOE];
  char   poltyp[8];

  SCKGETC("POLTYP", 1, 8, &actvals, poltyp);
  for (i = start_index; i < (n + start_index); i++)
    {
      l[i] = 0.;

      if (toupper(poltyp[0]) == 'L')
        fleg_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'C')
        fcheb_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'P')
        fpoly_2D (x[i], y[i], xp, ncoef);
      else
	printf("ERROR - You have tried an invalid polynom type \n"); 

      for (icoef = 1; icoef <= ncoef; icoef++)
	l[i] += coef[icoef] * xp[icoef];
    }
}

/* ------------------------------------------------------------------------- */
#ifdef __STDC__
  void mos_writedisp_2D(int line, int slit, int ypix, 
		      double y, int numrow, double rms) 
#else
  void mos_writedisp_2D(line, slit, ypix, y, numrow, rms)  
  int  line, slit, ypix, numrow;
  double y, rms;
#endif
{
  int icoef,iexp,kun,ibuff;
  double coef2D[MAXNCOE], poly;
  char descname[80];


  /* make 1D coeficients from 2D coeficients -- last 5 depend on (y-ystart)*/
  for (icoef = 1; icoef <= (maxcoef-(degy+degxy)); icoef++)
    coef2D[icoef-1] = coef[icoef];
  poly = y;
  for (iexp = 1; iexp <= degy; iexp++)
    {
      coef2D[0] = coef2D[0] + coef[(maxcoef-degy)-degxy+iexp]*poly;
      poly = poly * y;
    }
  poly = y;
  for (iexp = 1; iexp <= degxy; iexp++)
    {
      coef2D[1] = coef2D[1] + coef[maxcoef-degxy+iexp] * poly;
      poly = poly * y;
    }
  TCEWRI(tide, line, colslit, &slit);
  TCEWRI(tide, line, colline, &ypix);
  TCEWRD(tide, line, coly,    &y);
  TCEWRD(tide, line, colrms,  &rms); 
  
  /*      if (nbline<line) nbline=line;   */
  /*if (nbline<line) nbline=numrow;*/
  
  for (icoef=1; icoef<=maxcoef-(degy+degxy); icoef++)
    TCEWRD(tide, line, colcoef[icoef], &coef2D[icoef-1]);


  /* write 2D-fit parameters to table coerbr */
  ibuff = maxcoef - (degxy + degy);
  SCDWRI(tide, "DEGX", &ibuff, 1, 1, &kun);
  SCDWRI(tide, "DEGXY", &degxy, 1, 1, &kun);
  SCDWRI(tide, "DEGY", &degy, 1, 1, &kun);
  sprintf(descname,"coefslit%d",slit);
  SCDWRD(tide, descname, &coef[1], 1, maxcoef, &kun);
}









/*  ------------------------------------------------------------------------
 
    Here the long slit utilities modified for multi object spectroscopy by
    Sabine Moehler:

    ------------------------------------------------------------------------ */

#ifdef __STDC__
void 
mos_savedisp (double save[])
#else
void 
mos_savedisp (save)
     double save[];
#endif

{

  int i;

  /* printf("Dispersion Relation. Degree: %d. Refdeg: %d. MaxCoef:%d\n",
     fdeg,refdeg,maxcoef);
     printf("Coefficients: ");  */

  for (i = 0; i <= ncoef-1; i++)
    {
      save[i] = coef[i+1];
      /*     printf("save[%d] =  %f ",i, save[i]);  */
    }

}
/* ------------------------------------------------------------------------- */



#ifdef __STDC__
void 
printdisp (void)
#else
void 
printdisp ()
#endif

{

  int i;

  printf ("Dispersion Relation. Degree: %d. Refdeg: %d. MaxCoef:%d\n",
	  fdeg, refdeg, maxcoef);
  printf ("Coefficients: ");
  for (i = 1; i <= ncoef; i++)
    printf (" %f ", coef[i]);
  printf ("\n");

}
/* ------------------------------------------------------------------------- */



#ifdef __STDC__
void 
setrefdeg (int deg)
#else
void 
setrefdeg (deg)
     int deg;
#endif

{

  refdeg = deg;
  maxcoef = refdeg + 1;

}
/* ------------------------------------------------------------------------- */

#ifdef __STDC__
void 
set_zero (int deg)
#else
void 
set_zero (deg)
     int deg;
#endif


{

  int i;

  fdeg = deg;
  refdeg = deg;
  ncoef = deg + 1;
  maxcoef = deg + 1;

  for (i = 1; i <= ncoef; i++)
    coef[i] = 0;
}


/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void 
setdisp (int deg, double coefs[])
#else
void 
setdisp (deg, coefs)
     int deg;
     double coefs[];
#endif

{
  int i;

  fdeg = deg;
  refdeg = deg;
  ncoef = deg + 1;
  maxcoef = deg + 1;

  for (i = 1; i <= ncoef; i++)
    coef[i] = coefs[i - 1];

  FIT_SUCCESS = 1;

}



/* ------------------------------------------------------------------------- */


#ifdef __STDC__
double 
mos_fit_disp (int *ndata, int *deg, double x[],
	      double l[], double *chi)
#else
double 
mos_fit_disp (ndata, deg, x, l, chi)
     int *ndata, *deg;
     double x[], l[], *chi;
#endif

{

  double *sig, chisq, **covar;
  int    *lista;
  int    actvals, maxdeg, i; 
  char   poltyp[8];

/* printf("NData : %d   ---   Deg value : %d\n",*ndata,*deg); */

  refdeg = *deg; 
  maxdeg = *ndata - 1;
  if (refdeg > maxdeg)
    fdeg = maxdeg;
  else
    fdeg = refdeg;
  ncoef = fdeg + 1;
  maxcoef = refdeg + 1;

  if (*ndata < 2)
    {
      printf ("Not enough lines (minimum is 2). \nNo dispersion relation computed\n");
      return (-2.);
    }

  else if (fdeg < 1)
    {
      printf ("Degree : %d. No dispersion relation fitted\n", *deg);
      return (-1.);
    }

  else
    {
      covar = (double **) dmatrix (1, *ndata, 1, *ndata);
      /*      chisq = (double *) dvector (0, *ndata);*/
      sig = (double *) dvector (1, *ndata);


      lista = (int *) ivector (1, ncoef);

      for (i = 1; i <= ncoef; i++)
	lista[i] = 1;
      for (i = 1; i <= *ndata; i++)
        sig[i] = 1.;

      SCKGETC("POLTYP", 1, 8, &actvals, poltyp);
      if (toupper(poltyp[0]) == 'L')
        dlfit(x,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,dleg);
      else if (toupper(poltyp[0]) == 'C')
        dlfit(x,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,dcheb);
      else
	dlfit(x,l,sig,*ndata,coef,lista,ncoef,covar,&chisq,dpoly);
 
      *chi = chisq;

      free_dmatrix (covar, 1, *ndata, 1, *ndata);
      /*      free_dvector (chisq, 0, *ndata);*/
      free_dvector (sig, 1, *ndata);
      free_ivector (lista, 1, ncoef);

      pixbin = coef[2];
      return (pixbin);
    }
}

/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void 
mos_eval_disp (double x[], double l[], int n)
#else
void 
mos_eval_disp (x, l, n)
     double x[], l[];
     int n;
#endif

{

  int actvals, i, icoef;
  double xp[MAXNCOE];
  char   poltyp[8];

  SCKGETC("POLTYP", 1, 8, &actvals, poltyp);
  for (i = start_index; i < (n + start_index); i++)
    {
      l[i] = 0.;
      if (toupper(poltyp[0]) == 'L')
        dleg (x[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'C')
        dcheb (x[i], xp, ncoef);
      else
	dpoly (x[i], xp, ncoef);
      for (icoef = 1; icoef <= ncoef; icoef++)
	l[i] += coef[icoef] * xp[icoef];
    }
}
/* ------------------------------------------------------------------------- */

#ifdef __STDC__
void 
mos_initdisp (char name[], char mode[], int start)
#else
void 
mos_initdisp (name, mode, start)
     char name[], mode[];
     int start;
#endif

{

  int icoef;
  char colnam[30], numb[10];
  int actvals, null;
  int kunit;
  int ncol, nrow, nsort, allcol, allrow;


  start_index = start;

  if (toupper (mode[0]) == 'N')
    {				/*  NEW table */
      if (TCTINI (name, F_TRANS, F_IO_MODE, 5L, 10L, &tide))
	SCTPUT ("**** Error while creating output table");
      nbline = 0;
    }
  else
    {
      if (TCTOPN (name, F_IO_MODE, &tide))
	SCTPUT ("**** Error while opening output table");
      SCDRDD (tide, "LNPIX", 1L, 1L, &actvals, &pixbin, &kunit, &null);
      SCDRDI (tide, "LNDEG", 1L, 1L, &actvals, &refdeg, &kunit, &null);
      SCDRDI (tide, "LNCOE", 1L, 1L, &actvals, &maxcoef, &kunit, &null);
      /* fdeg=refdeg, ncoef=maxcoef; */
      TCIGET (tide, &ncol, &nrow, &nsort, &allcol, &allrow);
      nbline = nrow;
    }

  TCCSER (tide, ":SLIT", &colslit);
  if (colslit == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Slit Number",
	    "SLIT", &colslit);

  TCCSER (tide, ":ROW", &colline);
  if (colline == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Row Number",
	    "ROW", &colline);

  TCCSER (tide, ":Y", &coly);
  if (coly == (-1))
    TCCINI (tide, D_R8_FORMAT, 1, "F8.2", "Y Value",
	    "Y", &coly);

  TCCSER (tide, ":RMS", &colrms);
  if (colrms == (-1))
    TCCINI (tide, D_R8_FORMAT, 1, "F8.4", "Angstrom",
	    "RMS", &colrms);

  for (icoef = 1; icoef <= maxcoef; icoef++)
    {
      strcpy (colnam, ":COEF_");
      sprintf (numb, "%d", icoef);
      strcat (colnam, numb);
      TCCSER (tide, colnam, &colcoef[icoef]);
      if (colcoef[icoef] == (-1))
	TCCINI (tide, D_R8_FORMAT, 1, "F16.10", "Coefficients",
		colnam, &colcoef[icoef]);
    }
}
/* ------------------------------------------------------------------------- */

#ifdef __STDC__
void 
finishdisp (void)
#else
void 
finishdisp ()
#endif

{

  int kunit;

  SCDWRD (tide, "LNPIX", &pixbin, 1L, 1L, &kunit);
  SCDWRI (tide, "LNDEG", &refdeg, 1L, 1L, &kunit);
  SCDWRI (tide, "LNCOE", &maxcoef, 1L, 1L, &kunit);


  TCSINI(tide);
  TCTCLO (tide);

}

/*---------------------------------------------------------------------------*/
#ifdef __STDC__
int 
mos_readdisp (int y, int slit)
#else
int 
mos_readdisp (y, slit)
     int y, slit;
#endif

/* long int  y;     Pixel number of the row */
{

  int line, yval, linext, ydif, null, icoef;
  int mindif, actslit;

  mindif = -1;

  for (line = 1; line <= nbline; line++)
    {
      TCERDI (tide, line, colline, &yval, &null);
      TCERDI (tide, line, colslit, &actslit, &null);
      if (!null && actslit == slit)
	{
	  ydif = y - yval;
	  if (ydif < 0)
	    ydif = (-1) * ydif;
	  if (mindif < 0)
	    mindif = ydif;
	  if (ydif <= mindif)
	    mindif = ydif, linext = line;
	}
    }

  if (mindif == -1)
    return (-1);

  fdeg = refdeg, ncoef = maxcoef;

  for (icoef = 1; icoef <= ncoef; icoef++)
    TCERDD (tide, linext, colcoef[icoef], &coef[icoef], &null);

  return (0);

}


/*---------------------------------------------------------------------------

  mos_readdisp set's the coefficient to the next calibrated row, but this
  mean that no interpolation between the rows is possible!!!

  take mos_readdisp1 to set coeficients to:

  next lower calibrated row uplow = 0
  next upper calibrated row uplow = 1

  (for idim = 1)


  calculate the coeficients from the 2-dimensiona fit as stored in descriptors
  but not in table columns

  (for idim = 2)


  Thomas Szeifer, LSW Heidelberg, 12.07.98
*/


#ifdef __STDC__
int 
mos_readdisp1 (int y, int slit, int idim, int uplow, double *weight)
#else
int 
mos_readdisp1 (y, slit, idim, uplow, weight)
     int y, slit, idim, uplow;
     double *weight;
#endif

/* long int  y;     Pixel number of the row */
{

  int line, yval=0, linext, ydif, null, icoef, yval_last=0;
  int mindif, actslit;
  int ybuff;
  double ytmp,yprev;

  mindif = -1;
  *weight = 0;

  for (line = 1; line <= nbline; line++)
    {
      yprev = (double) yval;
      TCERDD (tide, line, colline, &ytmp, &null);
      yval = NINT(ytmp);
      TCERDI (tide, line, colslit, &actslit, &null);

      
      if (!null && actslit == slit)
	{
	  if (y < yval && yval_last == 0) /* below first y row calibrated */
	    {
	      mindif = 1;
	      ydif = y - yval;
	      
	      if (uplow == 0) /* read the first time */
 		{
		  linext = line;
		  TCERDD (tide, line, colline, &ytmp, &null);
		  *weight = (double) ytmp;
		}
	      else /* read the second time */
		{
		  linext = line+1;
		  TCERDD (tide, line+1, colline, &ytmp, &null);
		  *weight = (double) ytmp;
		}

	    }
	  else if (y >= yval_last && y < yval)
	    {
	      mindif = 1;
	      ydif = y - yval;

	      if (uplow == 0)
		{
		  linext = line - 1;
		  *weight = yprev;
		}
	      else
		{
		  linext = line;
		  *weight = (double) yval;
		}
	    }
	  else if (y >= yval_last && yval_last > 0)
	    {
	      mindif = 1;
	      ydif = y - yval;

	      if (uplow == 0)
		{
		  linext = line - 1;
		  *weight = yprev;
		}
	      else
		{
		  linext = line;
		  *weight = (double) yval;
		}
	    }

	  yval_last = yval;
	}
      
      /*     yval_last = yval;*/

    }

  if (mindif == -1)
    return (-1);

  fdeg = refdeg, ncoef = maxcoef;

  for (icoef = 1; icoef <= ncoef; icoef++)
    TCERDD (tide, linext, colcoef[icoef], &coef[icoef], &null);

  return (0);

}



/* ------------------------------------------------------------------------- */

#ifdef __STDC__
void 
mos_writedisp (int line, int slit, int ypix,
	       double y, int numrow, double rms)
#else
void 
mos_writedisp (line, slit, ypix, y, numrow, rms)
     int line, slit, ypix, numrow;
     double y, rms;
#endif

{

  int icoef;

  TCEWRI (tide, line, colslit, &slit);
  TCEWRI (tide, line, colline, &ypix);
  TCEWRD (tide, line, coly, &y);
  TCEWRD (tide, line, colrms, &rms);

/*      if (nbline<line) nbline=line;   */
  if (nbline < line)
    nbline = numrow;

  for (icoef = 1; icoef <= maxcoef; icoef++)
    TCEWRD (tide, line, colcoef[icoef], &coef[icoef]);
}













/* ---------------------------------------------------------------------------
       Non linear fitting utilities in two dimension   by Thomas Szeifert. 

       The functions:
       
       fleg_2D       (Legendre  polynoms)
       fcheb_2D      (Chebichev polynoms)
       fpoly_2D      ("normal"  polynoms)

       are specially designed for fits of the dispersion relation in 
       spectroscopy - we presume a small slope of the central wavelength and
       the first order dispersion coeficient and fit these in y-direction
       with "normal" polynoms of the order degy and degxy respectively.

-----------------------------------------------------------------------------*/


#ifdef __STDC__
  void fleg_2D( double x1, double x2, double pl[], int nl)
#else
  void fleg_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* Legendre polynom in x1 - polynom in y and one mixed polynom x1*x2*/
  int j;
  float twox,f2,f1,d;
  
  pl[1]=1.0;
  pl[2]=x1;
  if (nl > 2) 
    {
      twox=2.0*x1;  /*Legendre in x*/
      f2=x1;
      d=1.0;
      for (j=3;j<=nl-(degy+degxy);j++) 
	{
	  f1=d;
	  d += 1.0;
	  f2 += twox;
	  pl[j]=(f2*pl[j-1]-f1*pl[j-2])/d;
	}

      pl[nl-(degy+degxy)+1]=x2;  /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];
      
      pl[nl-degxy+1] = x1*x2;    /*mixed polynom*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }
}

#ifdef __STDC__
  void fcheb_2D( double x1, double x2, double pl[], int nl)
#else
  void fcheb_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* Chebichev polynom in x1 - polynom in y and one mixed polynom x1*x2*/
  int j;
  float twox;
  
  pl[1]=1.0;
  pl[2]=x1;
  if (nl > 2) 
    {
      twox=2.0*x1;  /*Chebichev in x*/
      for (j=3;j<=nl-(degy+degxy);j++) 
	pl[j]=twox*pl[j-1]-pl[j-2];
  
      pl[nl-(degy+degxy)+1]=x2;  /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];
      
      pl[nl-degxy+1] = x1*x2;    /*mixed polynom*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }
}

#ifdef __STDC__
  void fpoly_2D( double x1, double x2, double pl[], int nl)
#else
  void fpoly_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* standard polynom in x1 (nl) && in x2 (degy) and 
     mixed polynom x1*x2 (degxy) */
  int j;

  
  pl[1]=1.0;
  if (nl > 1) 
    {
      for (j=2;j<=nl-(degy+degxy);j++)  /*polynom in x*/
	pl[j]=x1*pl[j-1];

      pl[nl-(degy+degxy)+1] = x2;       /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];

      pl[nl-degxy+1] = x1*x2;           /*mixed polynom in x & y*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }
}





